"""Implementation of the reporter framework which can be used to report on the
progress or result of operations within utopya.
"""

import logging
import os
import sys
from collections import Counter, OrderedDict, deque
from datetime import datetime as dt
from datetime import timedelta
from functools import partial
from shutil import get_terminal_size as _get_terminal_size
from typing import Callable, Dict, List, Union

import numpy as np
import paramspace as psp

from .tools import TTY_COLS, format_time

log = logging.getLogger(__name__)


# -----------------------------------------------------------------------------


class ReportFormat:
    """A report format aggregates callables for a single report parser and
    potentially multiple report writers. As a whole, it contains all arguments
    needed to generate a certain kind of report.

    It is used in :py:class:`utopya.reporter.Reporter` and derived classes,
    which are the classes that actually implement the parsers and writers.
    """

    def __init__(
        self,
        *,
        parser: Callable,
        writers: List[Callable],
        min_report_intv: float = None,
    ):
        """Initializes a ReportFormat object, which gathers callables needed to
        create a report in a certain format.

        Args:
            parser (Callable): The parser method to use
            writers (List[Callable]): The writer method(s) to use
            min_report_intv (float, optional): The minimum report interval of
                reports in this format. Determines the time (in seconds) that
                needs to have passed before the next report will be emitted.
        """
        self.parser = parser
        self.writers = writers

        self._min_report_intv = None
        self.min_report_intv = min_report_intv

        self.num_reports = 0
        self.last_report = dt.fromtimestamp(0)  # waaay back

    @property
    def min_report_intv(self) -> Union[timedelta, None]:
        """Returns the minimum report interval, i.e. the time that needs to
        have passed between two reports.
        """
        return self._min_report_intv

    @min_report_intv.setter
    def min_report_intv(self, sec: float):
        """Set the minimum report interval, directly converting it into a
        timedelta value.
        """
        self._min_report_intv = timedelta(seconds=sec) if sec else None

    @property
    def reporting_blocked(self) -> bool:
        """Determines whether this ReportFormat may be blocked from emission,
        e.g. because of the minimum report interval not having passed yet.

        If no minimum report interval is given, will always return False.
        Otherwise checks if at least that interval has passed since the last
        report.
        """
        if not self.min_report_intv:
            return False

        return (dt.now() - self.last_report) < self.min_report_intv

    def report(
        self, *, force: bool = False, parser_kwargs: dict = None
    ) -> bool:
        """Parses and writes a report corresponding to the callables defined in
        this report format.

        Args:
            force (bool, optional): If True, will ignore the minimum report
                interval and always perform a report.
            parser_kwargs (dict, optional): Keyword arguments passed on to the
                parser

        Returns:
            bool: Whether a report was generated or not
        """
        if not force and self.reporting_blocked:
            # Do not report
            return False

        # Generate the report
        log.debug("Creating report using parser '%s' ...", self.parser)
        report = self.parser(
            report_no=self.num_reports,
            **(parser_kwargs if parser_kwargs else {}),
        )
        log.debug("Parser created report of length %d.", len(report))

        # Write the report
        for writer_name, writer in self.writers.items():
            log.debug("Writing report using writer '%s' ...", writer_name)
            writer(report)

        # Update counter and last report time
        self.num_reports += 1
        self.last_report = dt.now()

        return True


# -----------------------------------------------------------------------------


class Reporter:
    """The Reporter class holds general reporting capabilities.

    It needs to be subclassed in order to specialize its reporting functions.
    """

    def __init__(
        self,
        *,
        report_formats: Union[List[str], Dict[str, dict]] = None,
        default_format: str = None,
        report_dir: str = None,
        suppress_cr: bool = False,
    ):
        """Initialize the Reporter base class.

        Args:
            report_formats (Union[List[str], Dict[str, dict]], optional): The
                report formats to use with this reporter. If given as list
                of strings, the strings are the names of the report formats as
                well as those of the parsers; all other parameters are the
                defaults. If given as dict of dicts, the keys are the names of
                the formats and the inner dicts are the parameters to create
                report formats from.
            default_format (str, optional): The name of the default report
                format; if None is given, the .report method requires the name
                of a report format.
            report_dir (str, optional): if reporting to a file; this is the
                base directory that is reported to.
            suppress_cr (bool, optional): Whether to suppress carriage return
                characters in writers. This option is useful when the reporter
                is not the only class that writes to a stream.
        """

        super().__init__()

        # Property-managed attributes
        self._report_formats = dict()
        self._default_format = None
        self._suppress_cr = False

        # Ensure the report_formats argument is a dict, then register them
        if report_formats is None:
            report_formats = dict()

        elif isinstance(report_formats, (list, tuple)):
            report_formats = {f: dict() for f in report_formats}

        for name, params in report_formats.items():
            self.add_report_format(name, **params)

        # Set default report format
        if default_format:
            self.default_format = default_format

        # Store report directory
        if report_dir:
            self.report_dir = os.path.expanduser(str(report_dir))
        else:
            self.report_dir = None

        # Other attributes
        self.suppress_cr = suppress_cr  # NOTE writers need to implement this

        log.debug("Reporter.__init__ finished.")

    # Properties ..............................................................

    @property
    def report_formats(self) -> dict:
        """Returns the dict of ReportFormat objects."""
        return self._report_formats

    @property
    def default_format(self) -> Union[None, ReportFormat]:
        """Returns the default report format or None, if not set."""
        return self._default_format

    @default_format.setter
    def default_format(self, name: str):
        """Given the name of the report formats, set the default value."""
        if name is not None:
            self._default_format = self.report_formats[name]
            log.debug("Set default report format to '%s'.", name)
        else:
            self._default_format = None
            log.debug("Unset default report format.")

    @property
    def suppress_cr(self) -> bool:
        """Whether to suppress a carriage return. Objects using the reporter
        can set this property to communicate that they will be putting content
        into the stdout stream as well. The writers can check this property
        and adjust their behaviour accordingly.
        """
        return self._suppress_cr

    @suppress_cr.setter
    def suppress_cr(self, val: bool):
        """Set the suppress_cr property.

        When setting this to True the first time, a linebreak is issued in
        order to not overwrite any previously written lines that ended with
        a carriage return character.
        """
        if val and not self.suppress_cr:
            print("")

        self._suppress_cr = val

    # Public API ..............................................................

    def add_report_format(
        self,
        name: str,
        *,
        parser: str = None,
        write_to: Union[str, Dict[str, dict]] = "stdout",
        min_report_intv: float = None,
        rf_kwargs: dict = None,
        **parser_kwargs,
    ):
        """Add a report format to this reporter.

        Args:
            name (str): The name of this format
            parser (str, optional): The name of the parser; if not given, the
                name of the report format is assumed
            write_to (Union[str, Dict[str, dict]], optional): The name of the
                writer. If this is a dict of dict, the keys will be
                interpreted as the names of the writers and the nested dict as
                the ``**kwargs`` to the writer function.
            min_report_intv (float, optional): The minimum report interval (in
                seconds) for this report format
            rf_kwargs (dict, optional): Further kwargs to ReportFormat.__init__
            **parser_kwargs: The kwargs to the parser function

        Raises:
            ValueError: A report format with this `name` already exists
        """
        if name in self.report_formats:
            raise ValueError(
                f"A report format with the name '{name}' already exists."
            )

        # Get the parser and writer function
        parser = self._resolve_parser(
            parser if parser else name, **parser_kwargs
        )
        writers = self._resolve_writers(write_to)

        # Initialise the ReportFormat object with the parsers and writers
        rf = ReportFormat(
            parser=parser,
            writers=writers,
            min_report_intv=min_report_intv,
            **(rf_kwargs if rf_kwargs else {}),
        )

        self._report_formats[name] = rf
        log.debug(
            "Added report format '%s' to %s.", name, self.__class__.__name__
        )

    def report(self, report_format: str = None, **kwargs) -> bool:
        """Create a report with the given format; if none is given, the default
        format is used.

        Args:
            report_format (str, optional): The report format to use
            **kwargs: Passed on to the ReportFormat.report() call

        Returns:
            bool: Whether there was a report

        Raises:
            ValueError: If no default format was set and no report format name
                was given
        """
        # Get the report format to use
        if report_format is None:
            if self.default_format is None:
                raise ValueError(
                    "Either a default format needs to be set for this "
                    f"{self.__class__.__name__} or the name of the report "
                    "format needs to be supplied to the .report method."
                )

            rf = self.default_format

        else:
            rf = self.report_formats[report_format]

        # Delegate reporting to the ReportFormat class
        return rf.report(**kwargs)

    def parse_and_write(
        self,
        *,
        parser: Union[str, Callable],
        write_to: Union[str, Callable],
        **parser_kwargs,
    ):
        """This function allows to select a parser and writer explicitly.

        Args:
            parser (Union[str, Callable]): The parser method to use.
            write_to (Union[str, Callable]): The write method to use. Can also
                be a sequence of names and/or callables or a Dict. For allowed
                specification formats, see the ._resolve_writers method.
            **parser_kwargs: Passed to the parser, if given
        """
        parser = self._resolve_parser(parser, **parser_kwargs)

        # Parse the report
        report = parser()
        log.debug(
            "Parsed report using '%s', got string of length %d.",
            parser,
            len(report),
        )

        # Determine the writers and write
        writers = self._resolve_writers(write_to)

        for writer_name, writer in writers.items():
            writer(report)
            log.debug("Wrote report using %s .", writer_name)

    # Private methods .........................................................

    def _resolve_parser(
        self, parser: Union[str, Callable], **parser_kwargs
    ) -> Callable:
        """Given a string or a callable, returns the corresponding callable.

        Args:
            parser (Union[str, Callable]): If a callable is already given,
                returns that; otherwise looks for a parser method with the
                given name in the attributes of this class.
            **parser_kwargs: Arguments that should be passed to the parser.
                If given, a new function is created where these arguments are
                already included.

        Returns:
            Callable: The desired parser function

        Raises:
            ValueError: If no parser with the given name is available
        """
        if not callable(parser):
            # A name was given; try to resolve from attributes
            try:
                parser = getattr(self, "_parse_" + parser)
            except AttributeError as err:
                raise ValueError(
                    f"No parser named '{parser}' available in "
                    f"{self.__class__.__name__}!"
                ) from err

            log.debug("Resolved parser: %s", str(parser))

        # `parser` is now a callable.
        # May want to partially apply kwargs
        if parser_kwargs:
            log.debug("Binding `parser_kwargs` to parser method ...")
            parser = partial(parser, **parser_kwargs)

        return parser

    def _resolve_writers(self, write_to) -> Dict[str, Callable]:
        """Resolves the given argument to a list of callable writer functions.

        Args:
            write_to: a specification of the writers to use. Allows many
                different ways of specifying the writer functions, depending
                on the type of the argument:

                    - str: the name of the writer method of this reporter
                    - Callable: the writer function to use
                    - sequence of str and/or Callable: the names and/or
                      functions to use
                    - Dict[str, dict]: the names of the writer functions and
                      additional keyword arguments.

                If the type is wrong, will raise.

        Returns:
            Dict[str, Callable]: the writers (key: name, value: writer method)

        Raises:
            TypeError: Invalid ``write_to`` argument
            ValueError: A writer with that name was already added or a writer
                with the given name is not available.
        """

        def get_callable_name(c) -> str:
            """Returns the name of the callable by inspecting attributes"""
            if hasattr(c, "__name__"):
                return c.__name__
            # Does not have that attribute, e.g. because it is a partial func
            return str(c)

        # The target dict of callables
        writers = {}

        # First, need to bring the argument into a uniform structure.
        # This requires checking the many possible input formats.

        # -- Single callable: can directly return
        if callable(write_to):
            return {get_callable_name(write_to): write_to}

        # -- Single string: bring it into dict format
        elif isinstance(write_to, str):
            write_to = {write_to: {}}

        # -- list: move the callables to the writers dict; string items remain.
        elif isinstance(write_to, (list, tuple)):
            wt = {}
            for item in write_to:
                if callable(item):
                    # Check if already present
                    if item in writers.values():
                        raise ValueError(
                            "Given writer callable with name "
                            f"'{get_callable_name(item)}' was already added!"
                        )
                    writers[get_callable_name(item)] = item

                elif isinstance(item, str):
                    # Add an empty entry to the new write_to dict
                    wt[item] = dict()

                else:
                    raise TypeError(
                        f"One item of given `write_to` argument {write_to} "
                        f"of type {type(write_to)} was neither a string nor a "
                        "callable!"
                    )

            # Use the new write_to dict
            write_to = wt

        # Ensure that the format is a dict now
        if not isinstance(write_to, dict):
            raise TypeError(
                f"Invalid type {type(write_to)} for argument `write_to`!"
            )

        # Now populate the writers dict with the remaining str-specified funcs
        for writer_name, params in write_to.items():
            try:
                writer = getattr(self, "_write_to_" + writer_name)

            except AttributeError as err:
                raise ValueError(
                    "No writer named '{}' available in {}!"
                    "".format(writer_name, self.__class__.__name__)
                ) from err

            # If given, partially apply the params
            if params:
                writer = partial(writer, **params)

            # Store in dict of writers
            writers[writer_name] = writer
            log.debug("Added writer with name '%s'.", writer_name)

        return writers

    # Parser methods ..........................................................

    # None available in the base class

    # Writer methods ..........................................................

    def _write_to_stdout(self, s: str, *, flush: bool = True, **print_kws):
        """Writes the given string to stdout using the print function.

        Args:
            s (str): The string to write
            flush (bool, optional): Whether to flush directly; default: True
            **print_kws: Other print function keyword arguments
        """
        if self.suppress_cr and print_kws.get("end") == "\r":
            # Enforce line feed
            print_kws["end"] = "\n"

        print(s, flush=flush, **print_kws)

    def _write_to_stdout_noreturn(self, s: str, *, prepend="  "):
        """Writes to stdout without ending the line. Always flushes.

        Args:
            s (str): The string to write
            prepend (str, optional): Is prepended to the string; useful because
                the cursor might block this point of the terminal
            report_no (int, optional): accepted from ReportFormat call
        """
        if not self.suppress_cr:
            print(prepend + s, flush=True, end="\r")
        else:
            print(prepend + s, flush=True, end="\n")

    def _write_to_log(
        self, s: str, *, lvl: int = 10, skip_if_empty: bool = False
    ):
        """Writes the given string via the logging module.

        Args:
            s (str): The string to log
            lvl (int, optional): The level at which to log at; default is 10,
                corresponding to the ``DEBUG`` level
            skip_if_empty (bool, optional): Whether to skip writing if ``s`` is
                empty.
        """
        if not s and skip_if_empty:
            return
        log.log(lvl, s)

    def _write_to_file(
        self,
        s: str,
        *,
        path: str = "_report.txt",
        mode: str = "w",
        skip_if_empty: bool = False,
    ):
        """Writes the given string to a file

        Args:
            s (str): The string to write
            path (str, optional): The path to write it to; will be assumed
                relative to the ``report_dir`` attribute; if that is not
                given, ``path`` needs to be absolute. By default, assumes that
                there is a ``report_dir`` given.
            mode (str, optional): Writing mode of that file
            skip_if_empty (bool, optional): Whether to skip writing if ``s`` is
                empty.

        Raises:
            ValueError: If ``report_dir`` was not set and ``path`` is relative.
        """
        if not s and skip_if_empty:
            log.debug("Not writing to file because the given string is empty.")
            return

        # For given relative paths, join them to the report directory
        if not os.path.isabs(path):
            if not self.report_dir:
                raise ValueError(
                    "Need either an absolute `path` argument or initialize "
                    f"the {self.__class__.__name__} with the `report_dir` "
                    "argument such that `path` can be "
                    "interpreted relative to that directory."
                )

            path = os.path.join(self.report_dir, path)

        log.debug("Writing given string (length %d) to %s ...", len(s), path)

        with open(path, mode) as file:
            file.write(s)

        log.debug("Finished writing.")


# -----------------------------------------------------------------------------


class WorkerManagerReporter(Reporter):
    """This class specializes the base :py:class:`~utopya.reporter.Reporter`
    to report on the :py:class:`~utopya.workermanager.WorkerManager` state and
    its progress.
    """

    TTY_MARGIN = 4
    """Margin to use when writing to terminal"""

    PROGRESS_BAR_SYMBOLS = dict(
        finished="▓", active_progress="▒", active="░", space=" "
    )
    """Symbols to use in progress bar parser"""

    # .........................................................................

    def __init__(
        self,
        wm: "utopya.workermanager.WorkerManager",
        *,
        mv: "utopya.multiverse.Multiverse" = None,
        **reporter_kwargs,
    ):
        """Initialize the specialized reporter for the
        :py:class:`~utopya.workermanager.WorkerManager`.

        It is aware of the WorkerManager and may additionally have acces to the
        :py:class:`~utopya.multiverse.Multiverse` it is embedded in, which
        provides additional information to report parsers.

        Args:
            wm (utopya.workermanager.WorkerManager): The associated
                WorkerManager instance
            mv (utopya.multiverse.Multiverse, optional): The Multiverse this
                reporter is used in. If this is provided, it can be used in
                report parsers, e.g. to provide additional information on
                simulations.
            **reporter_kwargs: Passed on to parent method
        """
        super().__init__(**reporter_kwargs)

        # Make sure that formats 'while_working' and 'after_work' are available
        if "while_working" not in self.report_formats:
            log.debug(
                "No report format 'while_working' found; adding one "
                "because it is needed by the WorkerManager."
            )

            self.add_report_format(
                "while_working",
                parser="progress_bar",
                write_to="stdout_noreturn",
            )

        if "after_work" not in self.report_formats:
            log.debug(
                "No report format 'after_work' found; adding one "
                "because it is needed by the WorkerManager."
            )

            self.add_report_format(
                "after_work", parser="progress_bar", write_to="stdout_noreturn"
            )

        # Store the WorkerManager and associate it with this reporter
        self._wm = wm
        wm.reporter = self
        log.debug("Associated reporter with WorkerManager.")

        # Other attributes
        self.mv = mv
        self.runtimes = []
        self.exit_codes = Counter()
        self._eta_info = dict()

        log.debug("WorkerManagerReporter initialised.")

    @property
    def wm(self) -> "utopya.workermanager.WorkerManager":
        """Returns the associated
        :py:class:`~utopya.workermanager.WorkerManager`
        """
        return self._wm

    # Properties that extract info from the WorkerManager .....................

    @property
    def task_counters(self) -> OrderedDict:
        """Returns a dict of task counters containing the following entries:

        - ``total``: total number of registered WorkerManager tasks
        - ``active``: number of currently active tasks
        - ``finished``: number of finished tasks, *including* tasks that
          were stopped via a stop condition
        - ``stopped``: number of tasks for which stop conditions were
          fulfilled, see :ref:`stop_conds`
        """
        num = OrderedDict()
        num["total"] = self.wm.task_count
        num["active"] = len(self.wm.active_tasks)
        num["finished"] = self.wm.num_finished_tasks
        num["stopped"] = len(self.wm.stopped_tasks)
        return num

    @property
    def wm_progress(self) -> float:
        """The WorkerManager's progress, between 0 and 1."""
        cntr = self.task_counters

        if cntr["total"] == 0:
            return 0.0

        active_progress = self.wm_active_tasks_progress  # in [0, 1]
        return (
            cntr["finished"] / cntr["total"]
            + active_progress * cntr["active"] / cntr["total"]
        )

    @property
    def wm_active_tasks_progress(self) -> float:
        """The active tasks' progress.
        If there are no active tasks in the worker manager, returns 0.
        """
        progs = [t.progress for t in self._wm.active_tasks]
        if progs:
            return np.mean(progs)
        return 0.0

    @property
    def wm_elapsed(self) -> Union[timedelta, None]:
        """Seconds elapsed since start of working or None if not yet started"""
        times = self.wm.times

        if times["start_working"] is None:
            # Not started yet
            return None

        elif times["end_working"] is None:
            # Currently working: measure against now
            return dt.now() - times["start_working"]

        # Finished working: measure against end of work
        return times["end_working"] - times["start_working"]

    @property
    def wm_times(self) -> dict:
        """Return the characteristics of WorkerManager times. Calls
        :py:meth:`~utopya.reporter.WorkerManagerReporter.get_progress_info`
        without any additional arguments.
        """
        return self.get_progress_info()

    # Methods working on data .................................................

    def register_task(self, task: "utopya.task.WorkerTask"):
        """Given the task object, extracts and stores some information like
        its run time or its exit code.
        Exit codes are aggregated over multiple registrations.

        This can be used as a callback function from a WorkerTask object.

        Args:
            task (utopya.task.WorkerTask): The WorkerTask to extract
                information from.
        """
        if "run_time" in task.profiling:
            self.runtimes.append(task.profiling["run_time"])

        self.exit_codes[task.worker_status] += 1

    def calc_runtime_statistics(
        self, min_num: int = 10
    ) -> Union[OrderedDict, None]:
        """Calculates the current runtime statistics.

        Args:
            min_num (int, optional): Minimum number of runtimes that need to
                be registered for these statistics to actually be computed.
                If below this number, will return None.

        Returns:
            Union[OrderedDict, None]: The runtime statistics or None, if there
                were too few entries.
        """
        if not self.runtimes or len(self.runtimes) < min_num:
            return None

        # Throw out Nones and convert to np.array
        rts = np.array([rt for rt in self.runtimes if rt is not None])

        d = OrderedDict()
        d["total (CPU)"] = np.sum(rts)
        d["total (wall)"] = self.wm_elapsed.total_seconds()
        d["mean"] = np.mean(rts)
        d[" (last 50%)"] = np.mean(rts[-len(rts) // 2 :])
        d[" (last 20%)"] = np.mean(rts[-len(rts) // 5 :])
        d[" (last 5%)"] = np.mean(rts[-len(rts) // 20 :])
        d["std"] = np.std(rts)
        d["min"] = np.min(rts)
        d["at 25%"] = np.percentile(rts, 25)
        d["median"] = np.median(rts)
        d["at 75%"] = np.percentile(rts, 75)
        d["max"] = np.max(rts)

        return d

    def get_progress_info(self, **eta_options) -> Dict[str, float]:
        """Compiles a dict containing progress information for the current
        work session.

        Args:
            **eta_options: Passed on to method calculating ``est_left``,
                :py:meth:`._compute_est_left`.

        Returns:
            Dict[str, float]: Progress information. Guaranteed to contain the
                keys ``start``, ``now``, ``elapsed``, ``est_left``,
                ``est_end``, and ``end``.
        """
        d = dict(
            start=self.wm.times["start_working"],
            now=dt.now(),
            elapsed=self.wm_elapsed,
            est_left=None,
            est_end=None,
            end=self.wm.times["end_working"],
        )

        # Add estimate time remaining and ETA, if the WorkerManager started.
        if d["start"] is not None:
            progress = self.wm_progress
            if progress > 0.0:
                d["est_left"] = self._compute_est_left(
                    progress=progress, elapsed=d["elapsed"], **eta_options
                )

        if d["est_left"] is not None:
            d["est_end"] = d["now"] + d["est_left"]

        return d

    def _compute_est_left(
        self,
        *,
        progress: float,
        elapsed: timedelta,
        mode: str = "from_start",
        progress_buffer_size: int = 60,
    ) -> timedelta:
        """Computes the estimated time left until the end of the work session
        (ETA) using the current progress value and the elapsed time.
        Depending on ``mode``, additional information may be included in the
        calculation.

        Args:
            progress (float): The current progress value, in (0, 1]
            elapsed (datetime.timedelta): The elapsed time since start
            mode (str, optional): By which mode to calculate the ETA. Available
                modes are:

                    - ``from_start``, where ETA is computed from the start of
                        work session.
                    - ``from_buffer``, where ETA is computed from a more
                        recent point during the work session. This uses a
                        buffer to keep track of recent progress and computes
                        the ETA against the oldest record (controlled by
                        argument ``progress_buffer_size``), giving more
                        accurate estimates for long-running work sessions.

            progress_buffer_size (int, optional): The size of the ring buffer
                used in  ``from_buffer`` mode.

        Returns:
            datetime.timedelta: Estimate for how much time is left until the
                end of the work session.
        """
        if mode is None or mode == "from_start":
            return ((1.0 - progress) / progress) * elapsed

        elif mode == "from_buffer":
            # Get / set up the progress buffer: a circular buffer which holds
            # at most ``progress_buffer_size`` elements.
            # Each element is a (progress, elapsed) tuple.
            try:
                pbuf = self._eta_info["progress_buffer"]

            except KeyError:
                log.debug(
                    "Setting up progress buffer (maxlen: %d) ...",
                    progress_buffer_size,
                )
                pbuf = deque(
                    [(0.0, timedelta(0.0))], maxlen=progress_buffer_size
                )
                self._eta_info["progress_buffer"] = pbuf

            # Add new information to buffer
            pbuf.append((progress, elapsed))

            # Compute progress speed compared to first element of buffer
            _progress, _elapsed = pbuf[0]
            dp = progress - _progress

            if elapsed == _elapsed or dp <= 1.0e-16:
                # Buffer useless or too little progress made; use simpler mode
                return self._compute_est_left(
                    progress=progress, elapsed=elapsed, mode="from_start"
                )

            return ((1.0 - progress) / dp) * (elapsed - _elapsed)

        else:
            raise ValueError(
                f"Invalid ETA computation mode '{mode}'! "
                "Available modes: from_start, from_buffer"
            )

    # Parser methods ..........................................................

    # One-line parsers . . . . . . . . . . . . . . . . . . . . . . . . . . . .

    def _parse_task_counters(self, *, report_no: int = None) -> str:
        """Return a string that shows the task counters of the WorkerManager

        Args:
            report_no (int, optional): A counter variable passed by the
                :py:class:`~utopya.reporter.ReportFormat` call, indicating
                how often this parser was called so far.

        Returns:
            str: A str representation of the task counters of the WorkerManager
        """
        return ",  ".join([f"{k}: {v}" for k, v in self.task_counters.items()])

    def _parse_progress(self, *, report_no: int = None) -> str:
        """Returns a progress string

        Args:
            report_no (int, optional): A counter variable passed by the
                :py:class:`~utopya.reporter.ReportFormat` call, indicating
                how often this parser was called so far.

        Returns:
            str: A simple progress indicator
        """
        cntr = self.task_counters

        if cntr["total"] <= 0:
            return "(No tasks assigned to WorkerManager yet.)"

        return "Finished  {fin:>{digs:}d} / {tot:d}  ({p:.1f}%)" "".format(
            fin=cntr["finished"],
            tot=cntr["total"],
            digs=len(str(cntr["total"])),
            p=cntr["finished"] / cntr["total"] * 100,
        )

    def _parse_progress_bar(
        self,
        *,
        num_cols: Union[str, int] = "fixed",
        fstr: str = "  ╠{ticks[0]:}{ticks[1]:}{ticks[2]:}{ticks[3]:}╣ {info:}{times:}",
        info_fstr: str = "{total_progress:>5.1f}% ",
        show_times: bool = False,
        times_fstr: str = "| {elapsed:} elapsed | ~{est_left:} left ",
        times_fstr_final: str = "| finished in {elapsed:} ",
        times_kwargs: dict = {},
        report_no: int = None,
    ) -> str:
        """Returns a progress bar.

        It shows the amount of finished tasks, active tasks, and a percentage.

        Args:
            num_cols (Union[str, int], optional): The number of columns
                available for creating the progress bar. Can also be a string
                ``adaptive`` to poll terminal size upon each call, or ``fixed``
                to use the number of columns determined at import time.
            fstr (str, optional): The format string for the final output.
                Should contain the ``ticks`` 4-tuple, which makes up the
                progress bar, and can optionally contain the``info`` and
                ``times`` segments, formatted using the respective format
                string arguments.
            info_fstr (str, optional): The format string for the ``info``
                section of the final output. Available keys:

                    - ``total_progress``
                    - ``active_progress``
                    - ``cnt``, the task counters dictionary, see:
                        :py:meth:`~utopya.reporter.WorkerManagerReporter.task_counters`

            show_times (bool, optional): Whether to show a short version of the
                results of the times parser
            times_fstr (str, optional): Format string for times information
            times_fstr_final (str, optional): Format string for times
                information once the work session has ended
            times_kwargs (dict, optional): Passed on to ``times`` parser.
                Only used if ``show_times`` is set.
            report_no (int, optional): A counter variable passed by the
                :py:class:`~utopya.reporter.ReportFormat` call, indicating
                how often this parser was called so far.

        Returns:
            str: The one-line progress bar
        """
        # Get the task counter and check that some tasks have been assigned
        cntr = self.task_counters

        if cntr["total"] <= 0:
            return "(No tasks assigned to WorkerManager yet.)"

        # Determine the format string for the times
        if show_times:
            if self.wm_progress == 1.0:
                times_fstr = times_fstr_final
            times_str = self._parse_times(fstr=times_fstr, **times_kwargs)

        else:
            times_str = ""

        # Determine number of available columns
        if num_cols == "adaptive":
            num_cols = (
                _get_terminal_size((TTY_COLS, 20)).columns - self.TTY_MARGIN
            )

        elif num_cols == "fixed":
            num_cols = TTY_COLS - self.TTY_MARGIN

        # Get the active tasks' mean progress and calculate the total progress
        # (calling the wm_progress property would lead to inconsistencies)
        active_progress = self.wm_active_tasks_progress
        total_progress = (
            cntr["finished"] / cntr["total"]
            + active_progress * cntr["active"] / cntr["total"]
        )

        # Get the information string ready
        info_str = info_fstr.format(
            total_progress=total_progress * 100,
            active_progress=active_progress * 100,
            cnt=cntr,
        )

        # Determine the progress bar width
        pb_width = num_cols - (5 + len(info_str) + len(times_str))

        # Only return percentage indicator if the width would be _very_ short
        if pb_width < 4:
            return " {:>5.1f}% ".format(cntr["finished"] / cntr["total"] * 100)

        # Calculate the ticks for finished tasks
        ticks = dict()
        factor = pb_width / cntr["total"]
        ticks["finished"] = int(round(cntr["finished"] * factor))

        # Calculate the active ticks and those in progress
        # NOTE Important to round only one of the two, leads to artifacts
        #      otherwise
        ticks["active_progress"] = int(
            active_progress * cntr["active"] * factor
        )
        ticks["active"] = (
            int(round(cntr["active"] * factor)) - ticks["active_progress"]
        )

        # Calculate spaces from the sum of all of the above
        ticks["space"] = pb_width - sum(ticks.values())

        # Have all info now, let's go format!
        syms = self.PROGRESS_BAR_SYMBOLS
        return fstr.format(
            ticks=(
                syms["finished"] * ticks["finished"],
                syms["active_progress"] * ticks["active_progress"],
                syms["active"] * ticks["active"],
                syms["space"] * ticks["space"],
            ),
            info=info_str,
            times=times_str,
        )

    def _parse_times(
        self,
        *,
        fstr: str = "Elapsed:  {elapsed:<8s}  |  Est. left:  {est_left:<8s}  |  Est. end:  {est_end:<10s}",
        timefstr_short: str = "%H:%M:%S",
        timefstr_full: str = "%d.%m., %H:%M:%S",
        use_relative: bool = True,
        times: dict = None,
        report_no: int = None,
        **progress_info_kwargs,
    ) -> str:
        """Parses the WorkerManager's time information, including estimated
        time left or others.

        Args:
            fstr (str, optional): The main format string; gets as keys the
                results of the WorkerManager time information.
                Available keys: ``elapsed``, ``est_left``, ``est_end``,
                ``start``, ``now``, ``end``.
            timefstr_short (str, optional): A time format string for absolute
                dates; short version.
            timefstr_full (str, optional): A time format string for absolute
                dates; long (ideally: full) version.
            use_relative (bool, optional): Whether for a date difference of 1
                to use relative dates, e.g. ``Today, 13:37``.
            times (dict, optional): A dict of times to use; this is mainly
                for testing purposes!
            report_no (int, optional): The report number passed by ReportFormat
            **progress_info_kwargs: Passed on to method calculating progress
                :py:meth:`~utopya.reporter.WorkerManagerReporter.get_progress_info`

        Returns:
            str: A string representation of the time information
        """
        # If no explicit times were given, calculate them now
        if times is None:
            times = self.get_progress_info(**progress_info_kwargs)

        # The dict of strings that is filled and passed to the fstr
        tstrs = dict()

        # Convert some values to duration strings
        if times["elapsed"]:
            tstrs["elapsed"] = format_time(times["elapsed"])
        else:
            tstrs["elapsed"] = "(not started)"

        if times["est_left"]:
            tstrs["est_left"] = format_time(times["est_left"], max_num_parts=2)
        else:
            # No est_left available
            # Distinguish between finished and not started simulaltions
            if self.wm_progress == 1.0:
                tstrs["est_left"] = "(finished)"
            else:
                tstrs["est_left"] = "∞"

        # Check if the start and end times are given
        if not (times["start"] and times["est_end"]):
            # Not given -> not started yet
            tstrs["start"] = "(not started)"
            tstrs["now"] = times["now"].strftime(timefstr_full)
            tstrs["est_end"] = "(unknown)"
            tstrs["end"] = "(unknown)"

        else:
            # Were given.
            # Decide which time format string to use, depending on whether
            # start and end are on the same day, and whether to put a manual
            # prefix in front
            prefixes = dict()

            # Calculate timedelta in days
            delta_days = (times["est_end"].date() - times["start"].date()).days

            if delta_days == 0:
                # All on the same day -> use short format, no prefixes
                timefstr = timefstr_abs = timefstr_short

            elif delta_days == 1 and use_relative:
                # Use short format with prefixes
                timefstr = timefstr_short
                timefstr_abs = timefstr_full

                if times["now"].date() == times["start"].date():
                    # Same day as start -> end is tomorrow
                    prefixes["start"] = "Today, "
                    prefixes["est_end"] = "Tomorrow, "
                else:
                    # Same day as est end -> start was yesterday
                    prefixes["start"] = "Yesterday, "
                    prefixes["est_end"] = "Today, "

            else:
                # Full format
                timefstr = timefstr_abs = timefstr_full

            # Create the strings
            tstrs["start"] = times["start"].strftime(timefstr)
            tstrs["now"] = times["now"].strftime(timefstr_abs)
            tstrs["est_end"] = times["est_end"].strftime(timefstr)

            if times["end"]:
                tstrs["end"] = times["end"].strftime(timefstr_abs)
            else:
                tstrs["end"] = "(unknown)"

            # Add prefixes
            for key, prefix in prefixes.items():
                tstrs[key] = prefix + tstrs[key]

        return fstr.format(**tstrs)

    # Multi-line parsers . . . . . . . . . . . . . . . . . . . . . . . . . . .

    def _parse_runtime_stats(
        self,
        *,
        fstr: str = "  {k:<13s} {v:}",
        join_char="\n",
        ms_precision: int = 1,
        report_no: int = None,
    ) -> str:
        """Parses the runtime statistics dict into a multiline string

        Args:
            fstr (str, optional): The format string to use. Gets passed the
                keys ``k`` and ``v`` where ``k`` is the name of the entry and
                ``v`` its value. Note that ``v`` is a non-numeric value.
            join_char (str, optional): The join character / string to join the
                elements together.
            ms_precision (int, optional): Number of digits to represent the
                milliseconds part of the runtimes.
            report_no (int, optional): A counter variable passed by the
                :py:class:`~utopya.reporter.ReportFormat` call, indicating
                how often this parser was called so far.

        Returns:
            str: The multi-line runtime statistics
        """
        rtstats = self.calc_runtime_statistics()

        parts = [
            fstr.format(k=k, v=format_time(v, ms_precision=ms_precision))
            for k, v in rtstats.items()
        ]

        return join_char.join(parts)

    def _parse_report(
        self,
        *,
        fstr: str = "  {k:<{w:}s}  {v:}",
        min_num: int = 4,
        report_no: int = None,
        show_individual_runtimes: bool = True,
        task_label_singular: str = "task",
        task_label_plural: str = "tasks",
    ) -> str:
        """Parses a report for all tasks that were being worked on into a
        multiline string. The headings can be adjusted by keyword arguments.

        Args:
            fstr (str, optional): The format string to use. Gets passed the
                keys ``k`` and ``v`` where ``k`` is the name of the entry and
                ``v`` its value. Note that this format string is also used
                with ``v`` being a non-numeric value. Also, ``w`` can be used
                to have a key column of constant width.
            min_num (int, optional): The minimum number of universes needed to
                calculate runtime statistics.
            report_no (int, optional): A counter variable passed by the
                :py:class:`~utopya.reporter.ReportFormat` call, indicating
                how often this parser was called so far.
            show_individual_runtimes (bool, optional): Whether to report
                individual universe runtimes; default: True. This should be
                disabled if there are a huge number of universes.
            task_label_singular (str, optional): The label to use in the report
                when referring to a single task.
            task_label_plural (str, optional): The label to use in the report
                when referring to multiple tasks.

        Returns:
            str: The multi-line simulation report string
        """
        # List that contains the parts that will be written
        parts = []

        # Calculate the runtime statistics and add them to the parts
        rtstats = self.calc_runtime_statistics(min_num=min_num)

        if rtstats is not None:
            parts += ["Runtime Statistics"]
            parts += ["------------------"]
            parts += [""]
            parts += [
                "The statistics below are calculated from all "
                f"individual {task_label_singular} run times."
            ]
            parts += [""]
            parts += [
                "  # {}:  {} / {}".format(
                    task_label_plural, len(self.runtimes), len(self.wm.tasks)
                )
            ]
            parts += [""]
            parts += [
                fstr.format(k=k, v=format_time(v, ms_precision=1), w=12)
                for k, v in rtstats.items()
            ]
            parts += [""]
            parts += [""]

        # In cluster mode, add more information
        if self.wm.cluster_mode:
            _rcps = self.wm.resolved_cluster_params

            parts += ["Cluster Mode Information"]
            parts += ["------------------------"]
            parts += [""]
            parts += [
                fstr.format(k=k, v=_rcps[k], w=12)
                for k in (
                    "node_name",
                    "node_index",
                    "num_nodes",
                    "node_list",
                    "job_id",
                    "job_name",
                    "job_account",
                    "cluster_name",
                    "num_procs",
                )
                if _rcps.get(k) is not None
            ]
            parts += [""]
            parts += [""]

        # If stop conditions were fulfilled, inform about those
        if self.wm.stop_conditions:

            def task_names(sc: set) -> str:
                if not sc.fulfilled_for:
                    return "(None)"
                return ", ".join(sorted(t.name for t in sc.fulfilled_for))

            parts += ["Stop Conditions"]
            parts += ["---------------"]
            parts += [""]
            parts += [
                f"  {len(self.runtimes)} / {len(self.wm.tasks)} "
                f"{task_label_plural} were stopped due to at least one "
                "of the following stop conditions:"
            ]
            parts += [""]
            parts += [
                f"  {sc}\n" f"      {task_names(sc)}\n"
                for sc in self.wm.stop_conditions
            ]
            parts += [""]
            parts += [""]

        # Add individual universe run times
        if show_individual_runtimes:
            parts += [f"{task_label_singular.capitalize()} Runtimes"]
            parts += ["-" * len(parts[-1])]
            parts += [""]

            max_name_len = max([12] + [len(t.name) for t in self.wm.tasks])

            for task in self.wm.tasks:
                if "run_time" in task.profiling:
                    rt = task.profiling["run_time"]
                    parts += [
                        fstr.format(
                            k=task.name,
                            v=format_time(rt, ms_precision=1),
                            w=max_name_len,
                        )
                    ]

        return " \n".join(parts)

    def _parse_pspace_info(self, *, fstr: str, report_no: int = None) -> str:
        """Provides information about the parameter space.

        Extracts the ``parameter_space`` from the associated Multiverse's meta
        configuration and provides information on that.

        If there are multiple tasks specified, it is assumed that a sweep is or
        was being carried out and an information string is retrieved from the
        :py:class:`paramspace.paramspace.ParamSpace` object, which is then
        returned.
        If only a single task was defined, returns an empty string.

        Args:
            report_no (int, optional): A counter variable passed by the
                :py:class:`~utopya.reporter.ReportFormat` call, indicating
                how often this parser was called so far.

        Returns:
            str: If there is more than one task, returns the result of
                :py:meth:`paramspace.paramspace.ParamSpace.get_info_str`.
                If not, returns a string denoting that there was only one task.
        """
        if self.mv is None:
            raise ValueError(
                "No Multiverse associated with this reporter! "
                "Cannot parse ParamSpace information."
            )

        pspace = self.mv.meta_cfg["parameter_space"]
        if not isinstance(pspace, psp.ParamSpace):
            raise TypeError(f"Expected a ParamSpace object, got:\n\n{pspace}")

        if len(self.wm.tasks) <= 1:
            return ""

        return fstr.format(
            num_tasks=len(self.wm.tasks), sweep_info=pspace.get_info_str()
        )

    # Writer methods ..........................................................

    def _write_to_file(
        self,
        *args,
        path: str = "_report.txt",
        cluster_mode_path: str = "{0:}_{node_name:}{ext:}",
        **kwargs,
    ):
        """Overloads the parent method with capabilities needed in cluster mode

        All args and kwargs are passed through. If in cluster mode, the path
        is changed such that it includes the name of the node.

        Args:
            *args: Passed on to parent method
            path (str, optional): The path to save to
            cluster_mode_path (str, optional): The format string to use for the
                path in cluster mode. _Requires_ to contain the format key
                ``{0:}`` which retains the given ``path``, extension split off.
                Extension can be used via ``ext`` (already includes the dot).
                Additional format keys: ``node_name``, ``job_id``.
            **kwargs: Passed on to parent method
        """
        if not self.wm.cluster_mode:
            return super()._write_to_file(*args, path=path, **kwargs)

        # else: in cluster mode. Use the information to build a new path
        base_path, ext = os.path.splitext(path)
        fstr_args = [base_path]
        fstr_kwargs = dict(ext=ext)

        fstr_kwargs["node_name"] = self.wm.resolved_cluster_params["node_name"]
        fstr_kwargs["job_id"] = self.wm.resolved_cluster_params["job_id"]

        # Build the new path, then let the parent do the rest
        path = cluster_mode_path.format(*fstr_args, **fstr_kwargs)

        return super()._write_to_file(*args, path=path, **kwargs)



"""Tools that help testing models.

This mainly supplies the ModelTest class, which is a specialization of the
:py:class:`~utopya.model.Model` for usage in tests.
"""

import logging
import os

from .model import Model

# Get a logger
log = logging.getLogger(__name__)

# -----------------------------------------------------------------------------


class ModelTest(Model):
    """A class to use for testing Utopia models.

    It attaches to a certain model and makes it easy to load config files with
    which test should be carried out.
    """

    def __init__(
        self,
        model_name: str,
        *,
        test_file: str = None,
        use_tmpdir: bool = True,
        **kwargs,
    ):
        """Initialize the ModelTest class for the given model name.

        This is basically like the base class __init__ just that it sets the
        default value of ``use_tmpdir`` to True and renames TODO

        Args:
            model_name (str): Name of the model to test
            test_file (str, optional): The file this ModelTest is used in. If
                given, will look for config files relative to the folder this
                file is located in.
            use_tmpdir (bool, optional): Whether to use a temporary directory
                to write data to. The default value can be set here; but the
                flag can be overwritten in the create_mv and create_run_load
                methods. For false, the regular model output directory is used.

        Raises:
            ValueError: If the directory extracted from test_file is invalid
        """
        super().__init__(
            name=model_name,
            base_dir=(
                None if test_file is None else os.path.dirname(test_file)
            ),
            use_tmpdir=use_tmpdir,
            **kwargs,
        )



"""Defines arguments that are shared across various parts of the CLI"""

from typing import Union

import click


def add_options(options):
    def _add_options(func):
        for option in reversed(options):
            func = option(func)
        return func

    return _add_options


def default_none(ctx, _, value) -> Union[None, tuple]:
    """For use in ``multiple=True`` with ``nargs=-1`` where None is desired
    for a default value instead of an empty tuple.
    Pass this to the ``callback`` argument of an argument or an option.
    """
    if len(value) == 0:
        return None
    return value


# -----------------------------------------------------------------------------


INTERACTIVE_MODE_PROHIBITED_ARGS = (
    "run_cfg",
    "run_dir",
    "label",
    "set_params",
    "cluster_mode",
    "show_data_tree",
    "use_data_tree_cache",
    "load_parallel",
)
"""Argument names that may NOT be given in the interactive plotting prompt"""


# -----------------------------------------------------------------------------

OPTIONS = dict()

# -- Model selection
OPTIONS["label"] = (
    click.option(
        "--label",
        default=None,
        help=(
            "For model names that have multiple info bundles registered, a "
            "label is needed to unambiguously select the desired one. "
            "Alternatively, use the `utopya models set-default` CLI command "
            "to set a default label for a model."
        ),
    ),
)

OPTIONS["model_selection"] = (
    click.option(
        "--cfg-set",
        "--cs",
        default=None,
        help=(
            "If the chosen model provides configuration sets, use the config "
            "files from the chosen set for the run and plots config. "
            "Note that the specific arguments still take precedence over the "
            "values from the config sets; to use default paths, "
            "specify empty strings (`''`) for those arguments."
        ),
    ),
) + OPTIONS["label"]

# -- Universal flags
OPTIONS["debug_flag"] = (
    click.option(
        "-d",
        "--debug",
        count=True,
        help=(
            "The debug level."
            # TODO Expand
        ),
    ),
)
OPTIONS["cluster_mode"] = (
    click.option(
        "--cluster-mode",
        flag_value=True,
        default=None,
        help="Enables cluster mode.",
    ),
)

# -- Data loading options
OPTIONS["load"] = (
    click.option(
        "-P",
        "--load-parallel",
        flag_value=True,
        default=None,
        help=("If given, will force loading data in parallel."),
    ),
    click.option(
        "--use-data-tree-cache",
        "--tc",
        flag_value=True,
        default=None,
        help=(
            "If set, uses tree file caching: If no cache file exists, creates "
            "one after loading all data; if a tree file already exists, uses "
            "that to restore the data tree. This may bring a speed-up if the "
            "creation of the data tree takes a long time."
        ),
    ),
    click.option(
        "--show-data-tree",
        type=click.Choice(("full", "condensed", "none")),
        default="condensed",
        show_default=True,
        help="Controls which kind of data tree should be shown after loading.",
    ),
)


# -- Evaluation options
OPTIONS["eval"] = (
    click.option(
        "--plots-cfg",
        default=None,
        type=click.Path(exists=True, dir_okay=False, resolve_path=True),
        help=(
            "If given, uses the plots configuration file found at this path "
            "instead of the defaults for the model."
        ),
    ),
    click.option(
        "-u",
        "--update-plots-cfg",
        multiple=True,
        callback=default_none,
        help=(
            "Sets entries in the selected plots config. "
            "Example: -u my_plot.some_param=42 sets the 'some_param' entry "
            "in the plot configuration named 'my_plot'. "
            "Repeat the -u option to set multiple values."
        ),
    ),
    click.option(
        "--plot-only",
        "--po",
        "plot_only",
        multiple=True,
        callback=default_none,
        help=(
            "If given, will plot only those entries of the plot configuration "
            "that match the names given here. This can also be used to "
            "activate plots that are disabled in the specified plot "
            "configuration. Note that simple name globbing is supported, but "
            "the argument needs to be put into quotes to not conflict with "
            "the globbing done by the shell. "
            "Repeat the --po option to denote multiple plot_only arguments."
        ),
    ),
    click.option(
        "-i",
        "--interactive",
        flag_value=True,
        help=(
            "If set, the CLI will not exit after plotting finished, but allow "
            "to continue plotting in an interactive session. "
            "This option is useful for creating multiple plots in an "
            "iterative fashion, especially if data loading time is large. "
            "Note that all Multiverse-related configuration options can not "
            "be changed during the session and have to be set beforehand."
        ),
    ),
    click.option(
        "-R",
        "--reveal-output",
        flag_value=True,
        help="If set, opens the output directory after plotting finished.",
    ),
)



"""Test the command line methods"""

import argparse

import pytest

# import utopya.cltools as clt  # FIXME
import utopya.model_registry as mr
from utopya.cfg import load_from_cfg_dir
from utopya.yaml import write_yml

from ..test_cfg import tmp_cfg_dir
from ..test_model_registry import tmp_model_registry

# Fixtures --------------------------------------------------------------------


class MockArgs(dict):
    """An attribute dict to mock the CL arguments"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__dict__ = self


# -----------------------------------------------------------------------------


@pytest.mark.skip("Needs to be abstracted to not rely on CMake infrastructure")
def test_copy_model_files(capsys, monkeypatch):
    """This tests the copy_model_files CLI helper function. It only tests
    the dry_run because mocking the write functions would be too difficult
    here. The actual copying is tested
    """
    # First, without prompts
    copy_model_files = lambda **kws: clt.copy_model_files(
        **kws, use_prompts=False, skip_exts=[".pyc"], dry_run=True
    )

    # This should work
    copy_model_files(
        model_name="CopyMeBare", new_name="FooBar", target_project="Utopia"
    )

    # Make sure that some content is found in the output; this is a proxy for
    # the actual behaviour...
    out, _ = capsys.readouterr()
    print(out, "\n" + "#" * 79)

    # Path changes
    assert 0 < out.find("CopyMeBare.cc") < out.find("FooBar.cc")
    assert (
        0 < out.find("CopyMeBare/CopyMeBare.cc") < out.find("FooBar/FooBar.cc")
    )

    # Added the add_subdirectory command at the correct position
    assert (
        0
        < out.find("add_subdirectory(dummy)")
        < out.find("add_subdirectory(FooBar)")
        < out.find("add_subdirectory(HdfBench)")
    )

    # Without adding to CMakeLists.txt ...
    copy_model_files(
        model_name="CopyMeBare",
        add_to_cmakelists=False,
        new_name="FooBar2",
        target_project="Utopia",
    )
    out, _ = capsys.readouterr()
    print(out, "\n" + "#" * 79)
    assert not (0 < out.find("add_subdirectory(FooBar2)"))
    assert 0 < out.find("Remember to register the new model in the ")

    # These should not work due to bad model or project names
    with pytest.raises(ValueError, match="'dummy' is already registered!"):
        copy_model_files(
            model_name="CopyMeBare", new_name="dummy", target_project="Utopia"
        )
    _ = capsys.readouterr()

    with pytest.raises(ValueError, match="No Utopia project with name 'NoSu"):
        copy_model_files(
            model_name="CopyMeBare",
            new_name="FooBar",
            target_project="NoSuchProject",
        )
    _ = capsys.readouterr()

    # These should not work, because use_prompts == False
    with pytest.raises(ValueError, match="Missing new_name argument!"):
        copy_model_files(model_name="CopyMeBare")
    _ = capsys.readouterr()

    with pytest.raises(ValueError, match="Missing target_project argument!"):
        copy_model_files(model_name="CopyMeBare", new_name="FooBar")
    _ = capsys.readouterr()

    # Now, do it again, mocking some of the prompts
    copy_model_files = lambda **kws: clt.copy_model_files(**kws, dry_run=True)

    monkeypatch.setattr("builtins.input", lambda x: "MyNewModel")
    copy_model_files(model_name="CopyMeBare", target_project="Utopia")
    out, _ = capsys.readouterr()
    print(out, "\n" + "#" * 79)
    assert 0 < out.find("Name of the new model:      MyNewModel")

    monkeypatch.setattr("builtins.input", lambda x: "Utopia")
    copy_model_files(model_name="CopyMeBare", new_name="FooBar")
    out, _ = capsys.readouterr()
    print(out, "\n" + "#" * 79)
    assert 0 < out.find("Utopia project to copy to:  Utopia")

    monkeypatch.setattr("builtins.input", lambda x: "N")
    copy_model_files(
        model_name="CopyMeBare", new_name="FooBar", target_project="Utopia"
    )
    out, _ = capsys.readouterr()
    print(out, "\n" + "#" * 79)
    assert 0 < out.find("Not proceeding")

    def raise_KeyboardInterrupt(*_):
        raise KeyboardInterrupt()

    monkeypatch.setattr("builtins.input", raise_KeyboardInterrupt)
    copy_model_files(model_name="CopyMeBare", target_project="Utopia")
    out, _ = capsys.readouterr()
    print(out, "\n" + "#" * 79)

    monkeypatch.setattr("builtins.input", raise_KeyboardInterrupt)
    copy_model_files(model_name="CopyMeBare", new_name="FooBar")
    out, _ = capsys.readouterr()
    print(out, "\n" + "#" * 79)

    monkeypatch.setattr("builtins.input", raise_KeyboardInterrupt)
    copy_model_files(
        model_name="CopyMeBare", new_name="FooBar", target_project="Utopia"
    )
    out, _ = capsys.readouterr()
    print(out, "\n" + "#" * 79)


@pytest.mark.skip(reason="needs to be adapted")
def test_prompt_for_new_plot_args(capsys, monkeypatch):
    """Tests the prompt for new plot arguments"""
    # Some mock definition of the parser
    p_eval = argparse.ArgumentParser()
    p_eval.add_argument("model_name")
    p_eval.add_argument("run_dir_path", default=None, nargs="?")

    # plot configuration
    p_eval.add_argument("--plots-cfg", default=None, nargs="?")
    p_eval.add_argument("--update-plots-cfg", default=None, nargs="+")
    p_eval.add_argument("--plot-only", default=None, nargs="*")
    p_eval.add_argument("--interactive", action="store_true")
    p_eval.add_argument("--debug", action="store_true")

    p_eval.add_argument("--run-cfg-path", default=None, nargs="?")
    p_eval.add_argument("--set-cfg", default=None, nargs="+", type=str)
    p_eval.add_argument("--suppress-data-tree", action="store_true")
    p_eval.add_argument("--full-data-tree", action="store_true")
    p_eval.add_argument("--cluster-mode", action="store_true")
    # .........................................................................

    # The prompt function
    prompt = lambda **kws: clt.prompt_for_new_plot_args(**kws, parser=p_eval)

    # Specify mock argv and args
    argv = ["dummy", "--interactive", "--plot-only", "foo", "bar"]
    args = p_eval.parse_args(argv)

    # Specify some example parameters
    monkeypatch.setattr("builtins.input", lambda x: "--plot-only foo")
    argv, args = prompt(old_argv=argv, old_args=args)
    assert "--plot-only" in argv
    assert args.plot_only == ["foo"]
    assert not args.debug

    # Again, with some other input
    # NOTE Can't test the readline startup hook using monkeypatch.setattr, thus
    #      the full user input needs be specified ...
    monkeypatch.setattr(
        "builtins.input", lambda x: "--plot-only foo bar --debug"
    )
    argv, args = prompt(old_argv=argv, old_args=args)
    assert "--plot-only" in argv
    assert "--debug" in argv
    assert args.plot_only == ["foo", "bar"]
    assert args.debug

    # Some bad arguments (that are caught by the parser)
    with pytest.raises(SystemExit):
        monkeypatch.setattr("builtins.input", lambda x: "some pos. argument")
        prompt(old_argv=argv, old_args=args)

    # Some disallowed arguments (that are caught by the prompt function)
    with pytest.raises(ValueError, match="Disallowed arguments:"):
        monkeypatch.setattr(
            "builtins.input", lambda x: "some/run/cfg/path --plot-only foo"
        )
        prompt(old_argv=argv, old_args=args)



"""Tests `utopya batch` subcommand"""

import os
import shutil

import pytest

import utopya

from .. import get_cfg_fpath
from . import invoke_cli

BATCH_FILE = get_cfg_fpath("batch_file.yml")

# Fixtures --------------------------------------------------------------------
from .._fixtures import *


@pytest.fixture(autouse=True)
def register_test_project(tmp_projects):
    """Use on all tests in this module"""
    pass


# -----------------------------------------------------------------------------


def test_batch(with_test_models):  # FIXME creates test artifacts in output dir
    """Tests the `utopya batch` subcommand"""
    res = invoke_cli(("batch", "-d", "-s", "--note", "some_note", BATCH_FILE))
    print(res.output)

    # Currently, this will fail:
    assert res.exit_code != 0
    assert "Nothing to plot" in res.output



"""Implements utopya's PlotHelper specialization"""

import dantro.plot


class PlotHelper(dantro.plot.PlotHelper):
    """A specialization of dantro's
    :py:class:`~dantro.plot.plot_helper.PlotHelper` which is used for creating
    :py:mod:`matplotlib.pyplot`-based plots using
    :py:class:`~utopya.eval.plotcreators.PyPlotCreator`.

    This can be used to add additional helpers for use in utopya without
    requiring changes to dantro.

    .. note::

        The helpers implemented here should try to adhere to the interface
        exemplified by the dantro
        :py:class:`~dantro.plot.plot_helper.PlotHelper`, with the aim that they
        can then be migrated into dantro in the long run.
    """

    # .. Helper methods .......................................................
    # Can add helper methods here, prefixed with _hlpr_



"""Implements specialized data group classes that are based on group types
provided in :py:mod:`dantro.groups`.
Here, they are configured using their class variables.
"""

import logging

import dantro
import dantro.groups

from .containers import XarrayDC as _XarrayDC

log = logging.getLogger(__name__)

# -----------------------------------------------------------------------------


class UniverseGroup(dantro.groups.ParamSpaceStateGroup):
    """This group represents the data of a single universe"""


class MultiverseGroup(dantro.groups.ParamSpaceGroup):
    """This group is meant to manage the ``multiverse`` group of the loaded
    data, i.e. the group where output of the individual
    :py:class:`~utopya.eval.groups.UniverseGroup` objects is stored in.

    Its main aim is to provide easy access to universes. By default, universes
    are only identified by their ID, which is a zero-padded *string*. This
    group adds the ability to access them via integer indices.

    Furthermore, via dantro, an easy data selector method is available, see
    :py:meth:`dantro.groups.psp.ParamSpaceGroup.select`.
    """

    _NEW_GROUP_CLS = UniverseGroup


class TimeSeriesGroup(dantro.groups.TimeSeriesGroup):
    """This group is meant to manage time series data, with the container names
    being interpreted as the time coordinate.
    """

    _NEW_CONTAINER_CLS = _XarrayDC


class HeterogeneousTimeSeriesGroup(dantro.groups.HeterogeneousTimeSeriesGroup):
    """This group is meant to manage time series data, with the container names
    being interpreted as the time coordinate.
    """

    _NEW_CONTAINER_CLS = _XarrayDC


class GraphGroup(dantro.groups.GraphGroup):
    """This group is meant to manage graph data and create a NetworkX graph
    from it.
    """

    # Let new groups contain only time series
    _NEW_GROUP_CLS = TimeSeriesGroup

    # Define allowed member container types
    _ALLOWED_CONT_TYPES = (TimeSeriesGroup, _XarrayDC)

    # Expected names for the containers that hold vertex/edge information
    _GG_node_container = "_vertices"
    _GG_edge_container = "_edges"

    # _group_ attribute names for optionally providing additional information
    # on the graph.
    _GG_attr_directed = "is_directed"
    _GG_attr_parallel = "allows_parallel"
    _GG_attr_edge_container_is_transposed = "edge_container_is_transposed"
    _GG_attr_keep_dim = "keep_dim"

    # Whether warning is raised upon bad alignment of property data
    _GG_WARN_UPON_BAD_ALIGN = True



"""Implements batch running and evaluation of simulations"""

import logging
import os
import time
from copy import deepcopy as _deepcopy
from shutil import copy2 as _copy2
from typing import Callable, Dict, Sequence, Tuple, Union

from pkg_resources import resource_filename as _resource_filename

from .cfg import UTOPYA_CFG_FILE_PATHS as _UTOPYA_CFG_FILE_PATHS
from .cfg import load_from_cfg_dir as _load_from_cfg_dir
from .reporter import WorkerManagerReporter
from .task import MPProcessTask
from .tools import recursive_update as _recursive_update
from .workermanager import WorkerManager
from .yaml import load_yml as _load_yml
from .yaml import write_yml as _write_yml

_BTM_BASE_CFG_PATH = _resource_filename("utopya", "cfg/btm_cfg.yml")
"""Base configuration path of the batch task manager"""

_BTM_BASE_CFG = _load_yml(_BTM_BASE_CFG_PATH)
"""Actual base configuration of the batch task manager"""

_BTM_USER_DEFAULTS = _load_from_cfg_dir("batch")
"""User defaults for the batch task manager"""

_BTM_DEFAULTS = _recursive_update(
    _deepcopy(_BTM_BASE_CFG), _deepcopy(_BTM_USER_DEFAULTS)
)
"""Aggregated and recursively updated default batch task manager config"""

INVALID_TASK_NAME_CHARS = (
    "/",
    ":",
    ".",
    "?",
    "*",
)
"""Substrings that may not appear in task names"""

log = logging.getLogger(__name__)

# -----------------------------------------------------------------------------
# .. Definition of multiprocessing.Process target callables ...................
# These need to happen here because they need to be importable from the module
# and not be local function definitions.


def _eval_task(
    *,
    task_name: str,
    _batch_name: str,
    _batch_dirs: dict,
    _task_cfg_path: str,
    _create_symlinks: bool,
    model_name: str,
    model_kwargs: dict = {},
    use_data_tree_cache: bool = None,
    print_tree: Union[bool, str] = "condensed",
    plot_only: Sequence[str] = None,
    plots_cfg: str = None,
    update_plots_cfg: dict = {},
    **frozen_mv_kwargs,
):
    """The evaluation task target for the multiprocessing.Process. It sets up
    a :py:class:`utopya.model.Model`, loads the data, and performs plots.
    """
    import os

    from .model import Model

    log.hilight(
        "Setting up evaluation task '%s' for model '%s' ...",
        task_name,
        model_name,
    )

    model = Model(name=model_name, **model_kwargs)
    mv = model.create_frozen_mv(**frozen_mv_kwargs)

    # Create symlinks to improve crosslinking between related files/directories
    if _create_symlinks:
        log.progress("Creating symlinks ...")

        # ... back to the run directory, removing an existing one.
        _dst = os.path.join(mv.dm.dirs["out"], "_run")
        if os.path.islink(_dst):
            os.remove(_dst)
        os.symlink(mv.dm.dirs["data"], _dst)
        log.note("...to original run directory.")

        # ... from the evaluation output directory of the original run to the
        # evaluation data directory of this batch eval task.
        _dst = os.path.join(
            mv.dm.dirs["data"], "eval", f"{_batch_name}_eval_{task_name}"
        )
        os.symlink(mv.dm.dirs["out"], _dst)
        log.note(
            "...from original run directory back to this evaluation "
            "output directory."
        )

        # ... back to the task configuration file, removing an existing one.
        _dst = os.path.join(mv.dm.dirs["out"], "_task_cfg.yml")
        if os.path.islink(_dst):
            os.remove(_dst)
        os.symlink(_task_cfg_path, _dst)
        log.note("...to batch task configuration.")

        # With output directories that are *outside* of the batch run
        # directory, add links back and forth between them:
        if not mv.dm.dirs["out"].startswith(_batch_dirs["eval"]):
            # ... back to the batch run directory
            _dst = os.path.join(mv.dm.dirs["out"], "_batch_run")
            if os.path.islink(_dst):
                os.remove(_dst)
            os.symlink(_batch_dirs["batch_run"], _dst)
            log.note("...to batch run directory.")

            # ... from default eval directory to custom one
            _dst = os.path.join(_batch_dirs["eval"], task_name)
            os.symlink(mv.dm.dirs["out"], _dst)
            log.note(
                "...from batch evaluation output directory to custom "
                "output directory."
            )

    print("")

    # Load the data tree, potentially restoring from tree cache
    if not use_data_tree_cache:
        mv.dm.load_from_cfg()

    else:
        if not mv.dm.tree_cache_exists:
            mv.dm.load_from_cfg()
            mv.dm.dump()
        else:
            log.hilight("Restoring tree from cache file ...")
            mv.dm.restore()

    # Done loading
    if print_tree == "condensed":
        print(mv.dm.tree_condensed)

    elif print_tree:
        print(mv.dm.tree)

    # Start plotting ...
    mv.pm.plot_from_cfg(
        plots_cfg=plots_cfg, plot_only=plot_only, **update_plots_cfg
    )


# -----------------------------------------------------------------------------


class BatchTaskManager:
    """A manager for batch tasks"""

    RUN_DIR_TIME_FSTR = "%y%m%d-%H%M%S"
    """The time format string for the run directory"""

    # .........................................................................

    def __init__(self, *, batch_cfg_path: str = None, **update_batch_cfg):
        """Sets up a BatchTaskManager.

        Args:
            batch_cfg_path (str, optional): The batch file with all the task
                definitions.
            **update_batch_cfg: Additional arguments that are used to update
                the batch configuration.

        Raises:
            NotImplementedError: If ``run_tasks`` or ``cluster_mode`` were set
                in the batch configuration.
        """
        log.progress("Initializing BatchTaskManager ...")

        self._cfg = self._setup_batch_cfg(batch_cfg_path, **update_batch_cfg)
        self._timestamp_str = time.strftime(self.RUN_DIR_TIME_FSTR)

        self._dirs, self._name = self._setup_dirs(**self._cfg["paths"])
        log.note("Batch run directory:\n  %s", self.dirs["batch_run"])

        self._perform_backup(
            base_cfg=_BTM_BASE_CFG_PATH,
            user_cfg=_UTOPYA_CFG_FILE_PATHS.get("batch"),
            batch_file=batch_cfg_path,
            update_cfg=update_batch_cfg,
            batch_cfg=self._cfg,
        )

        # Some features are not yet implemented ...
        if self._cfg["cluster_mode"]:
            raise NotImplementedError("Cluster mode is not supported yet!")

        # Set up the WorkerManager and its reporter
        self._wm = WorkerManager(**self._cfg["worker_manager"])
        self._reporter = WorkerManagerReporter(
            self._wm,
            report_dir=self.dirs["batch_run"],
            **self._cfg["reporter"],
        )

        log.progress("Initialized BatchTaskManager.")
        log.note("  Parallelization level:     %s", self.parallelization_level)
        log.note("  Debug mode?                %s", self.debug)

    def __str__(self) -> str:
        return f"<BatchTaskManager '{self._name}'>"

    # .........................................................................

    @property
    def debug(self) -> bool:
        """Whether debug mode was enabled."""
        return self._cfg["debug"]

    @property
    def parallelization_level(self) -> str:
        return self._cfg["parallelization_level"]

    @property
    def run_defaults(self) -> dict:
        """A deepcopy of the run task defaults"""
        return _deepcopy(self._cfg["task_defaults"]["run"])

    @property
    def eval_defaults(self) -> dict:
        """A deepcopy of the eval task defaults"""
        return _deepcopy(self._cfg["task_defaults"]["eval"])

    @property
    def dirs(self) -> dict:
        """The directories associated with this BatchTaskManager"""
        return self._dirs

    # .........................................................................

    def perform_tasks(self):
        """Perform all run and eval tasks."""
        n_run = self._add_tasks(
            tasks=self._cfg["tasks"]["run"],
            defaults=self.run_defaults,
            add_task=self._add_run_task,
        )
        n_eval = self._add_tasks(
            tasks=self._cfg["tasks"]["eval"],
            defaults=self.eval_defaults,
            add_task=self._add_eval_task,
        )

        self._wm.start_working()
        log.success(
            "Finished %d run tasks and %d evaluation tasks.", n_run, n_eval
        )

    # .........................................................................

    @staticmethod
    def _setup_batch_cfg(batch_cfg_path: str, **update_batch_cfg) -> dict:
        """Sets up the BatchTaskManager configuration"""
        # Update defaults with configuration from file, if given
        batch_cfg = _load_yml(batch_cfg_path) if batch_cfg_path else {}
        batch_cfg = _recursive_update(_deepcopy(_BTM_DEFAULTS), batch_cfg)

        # Update again
        batch_cfg = _recursive_update(batch_cfg, _deepcopy(update_batch_cfg))

        # For debug mode, let the WorkerManager raise directly rather than only
        # issuing warnings (default).
        if batch_cfg["debug"]:
            batch_cfg["worker_manager"]["nonzero_exit_handling"] = "raise"

        # Evaluate parallelization level
        plevel = batch_cfg["parallelization_level"]

        if plevel == "task":
            batch_cfg["worker_manager"]["num_workers"] = 1
            batch_cfg["worker_kwargs"]["forward_streams"] = True

        elif plevel == "batch":
            task_defaults = batch_cfg["task_defaults"]

            task_defaults["run"] = _recursive_update(
                task_defaults["run"],
                dict(
                    worker_manager=dict(num_workers=1),
                    parameter_space=dict(
                        parallel_execution=dict(enabled=False)
                    ),
                ),
            )

        else:
            raise ValueError(
                f"Invalid parallelization_level '{plevel}'! "
                "Valid options are:  batch, task"
            )

        log.info("Loaded batch configuration.")
        return batch_cfg

    def _setup_dirs(
        self, out_dir: str, note: str = None
    ) -> Tuple[Dict[str, str], str]:
        """Sets up directories"""
        out_dir = os.path.expanduser(out_dir)

        if not os.path.isabs(out_dir):
            raise ValueError(
                f"Batch output directory needs to be absolute! Was: {out_dir}"
            )

        # Batch run directory
        fstr = "{timestamp:}"
        if note:
            fstr += "_{note:}"
        batch_run_dir_name = fstr.format(
            timestamp=self._timestamp_str, note=note
        )
        batch_run_dir = os.path.join(out_dir, batch_run_dir_name)

        # Create batch run directory
        dirs = dict(out=out_dir, batch_run=batch_run_dir)
        os.makedirs(batch_run_dir)

        # Subdirectories
        subdirs = (
            "config",
            "config/tasks",
            "eval",
            "logs",
        )
        for subdir in subdirs:
            dirs[subdir] = os.path.join(batch_run_dir, subdir)
            os.makedirs(dirs[subdir])

        return dirs, batch_run_dir_name

    def _perform_backup(self, **parts):
        """Stores the given configuration parts in the config directory"""
        log.info("Performing backups ...")
        cfg_dir = self.dirs["config"]

        for part_name, val in parts.items():
            _path = os.path.join(cfg_dir, part_name + ".yml")

            # If string, assume it's a path and store original file, which
            # preserves the comments and structure etc. Otherwise store the
            # (most probably dict-like) data as a YAML dump.
            if isinstance(val, str):
                if not os.path.exists(val):
                    continue
                log.debug("Copying %s config ...", part_name)
                _copy2(val, _path)

            elif not val:
                log.debug("'%s' was empty, nothing to back up.", part_name)

            else:
                log.debug("Dumping %s config dict ...", part_name)
                _write_yml(val, path=_path)

        log.note("  Backed up all involved configuration files.")

    # .........................................................................

    def _add_tasks(
        self, tasks: dict, defaults: dict, add_task: Callable
    ) -> int:
        """Adds all configured run tasks to the WorkerManager's task queue"""
        tasks = _deepcopy(tasks)

        for task_name, task_cfg in tasks.items():
            if any(s in task_name for s in INVALID_TASK_NAME_CHARS):
                raise ValueError(
                    f"Invalid task name '{task_name}'! May not contain any of "
                    "the following characters or substrings:  "
                    + " ".join(INVALID_TASK_NAME_CHARS)
                )

            task_cfg = _recursive_update(_deepcopy(defaults), task_cfg)
            add_task(task_name, **task_cfg)

        return len(tasks)

    def _add_run_task(self, name: str, **_):
        """Adds a single run task to the WorkerManager"""
        raise NotImplementedError(
            "Run tasks are not supported yet! Remove the corresponding "
            "entries from the BatchTaskManager configuration."
        )

    def _add_eval_task(
        self,
        name: str,
        *,
        model_name: str,
        out_dir: str,
        enabled: bool = True,
        priority: int = None,
        create_symlinks: bool = False,
        **eval_task_kwargs,
    ):
        """Adds a single evaluation task to the WorkerManager.

        Args:
            name (str): Name of this task
            model_name (str): Model name; required in task, thus already
                requiring it here.
            out_dir (str): The path to the data output directory, i.e. the
                directory where all plots will ned up in.
                This may be a format string containing any of the following
                keys: ``task_name``, ``model_name``, ``timestamp``,
                ``batch_name`` (combination of ``timestamp`` and the note).
                Relative paths are evaluated relative to the ``eval`` batch
                run directory.
            enabled (bool, optional): If False, will *not* add this task.
            priority (int, optional): Task priority; tasks with smaller value
                will be picked first.
            create_symlinks (bool, optional): Whether to create symlinks that
                add crosslinks between related directories, e.g.: from the
                output directory, link back to the task configuration; from the
                evaluation output directory alongside the simulation data, link
                to the batch output directory
            **eval_task_kwargs: All further evaluation task arguments.
        """

        def setup_eval_task(
            worker_kwargs: dict,
            model_name: str,
            out_dir: str,
            create_symlinks: bool,
            eval_task_kwargs: dict,
        ) -> dict:
            """Run before the task starts; sets up all arguments for it ..."""
            # Prepare the DataManager's output directory path
            out_dir = out_dir.format(
                task_name=name,
                model_name=model_name,
                timestamp=self._timestamp_str,
                batch_name=self._name,
            )
            out_dir = os.path.expanduser(out_dir)

            if not os.path.isabs(out_dir):
                out_dir = os.path.join(self.dirs["eval"], out_dir)

            # Update the task arguments accordingly, setting the DataManager's
            # output directory and adding metadata
            eval_task_kwargs = _recursive_update(
                eval_task_kwargs,
                dict(
                    _batch_name=self._name,
                    task_name=name,
                    model_name=model_name,
                    data_manager=dict(out_dir=out_dir),
                ),
            )

            # Backup the configuration, such that the task can create a symlink
            # back to it. This needs to be done by the task, because the
            # DataManager will create the output directory and will fail if it
            # already exists (which is good).
            task_cfg_path = os.path.join(
                self.dirs["config/tasks"], f"eval_{name}.yml"
            )
            _write_yml(eval_task_kwargs, path=task_cfg_path)
            eval_task_kwargs["_batch_dirs"] = self.dirs
            eval_task_kwargs["_task_cfg_path"] = task_cfg_path
            eval_task_kwargs["_create_symlinks"] = create_symlinks

            # Generate a new worker_kwargs dict, carrying over the given ones
            worker_kwargs = dict(
                args=(_eval_task,),
                read_stdout=True,
                stdout_parser="yaml_dict",
                **worker_kwargs,
            )
            worker_kwargs["popen_kwargs"]["kwargs"] = eval_task_kwargs

            # If the streams are to be saved, save them to the logs output
            # directory inside the batch directory.
            if worker_kwargs.get("save_streams", True):
                worker_kwargs["save_streams_to"] = os.path.join(
                    self.dirs["logs"], f"eval_{name}_" + "{name:}.log"
                )

            return worker_kwargs

        if not enabled:
            log.debug("Task '%s' was not enabled. Skipping.", name)
            return

        # Determine setup and worker kwargs
        worker_kwargs = _deepcopy(self._cfg["worker_kwargs"])
        setup_kwargs = dict(
            model_name=model_name,
            out_dir=out_dir,
            eval_task_kwargs=eval_task_kwargs,
            create_symlinks=create_symlinks,
        )

        # Add the task
        self._wm.add_task(
            TaskCls=MPProcessTask,
            name=name,
            priority=priority,
            setup_func=setup_eval_task,
            setup_kwargs=setup_kwargs,
            worker_kwargs=worker_kwargs,
        )



"""Implements seaborn-based plotting functions"""

import copy
import logging
from typing import Hashable, List, Sequence, Tuple, Union

import pandas as pd
import seaborn as sns
import xarray as xr
from dantro.exceptions import PlottingError
from dantro.plot.funcs.generic import (
    determine_encoding,
    figure_leak_prevention,
    make_facet_grid_plot,
)

from utopya.plotting import PlotHelper, is_plot_func

log = logging.getLogger(__name__)


# .. Seaborn's figure-level plot functions ....................................
SNS_PLOT_FUNCS = {
    "relplot": sns.relplot,
    "displot": sns.displot,
    "catplot": sns.catplot,
    "lmplot": sns.lmplot,
    "clustermap": sns.clustermap,
    "pairplot": sns.pairplot,
    "jointplot": sns.jointplot,
}

SNS_FACETGRID_KINDS = (
    "relplot",
    "displot",
    "catplot",
    "lmplot",
)

# .. Encodings for seaborn's figure-level plot functions ......................
# TODO Check if all are correct
SNS_ENCODINGS = {
    # FacetGrid: Distributions
    "displot": ("col", "row", "hue"),
    "catplot": ("y", "hue", "col", "row"),
    # FacetGrid: Relational
    "relplot": ("x", "y", "hue", "col", "row", "style", "size"),
    "lmplot": ("x", "y", "hue", "col", "row"),
    # Others
    "clustermap": ("hue", "col", "row"),
    "pairplot": ("hue",),
    "jointplot": (
        "x",
        "y",
        "hue",
    ),
}


# -----------------------------------------------------------------------------


@is_plot_func(use_dag=True, required_dag_tags=("data",))
def snsplot(
    *,
    data: dict,
    hlpr: PlotHelper,
    sns_kind: str,
    free_indices: Tuple[str],
    optional_free_indices: Tuple[str] = (),
    auto_encoding: Union[bool, dict] = None,
    reset_index: bool = False,
    to_dataframe_kwargs: dict = None,
    dropna: bool = False,
    dropna_kwargs: dict = None,
    sample: Union[bool, int] = False,
    sample_kwargs: dict = None,
    **plot_kwargs,
) -> None:
    """Interface to seaborn's figure-level plot functions.

    Plots on a facet grid:
        * ``relplot``:      seaborn.pydata.org/generated/seaborn.relplot.html
        * ``displot``:      seaborn.pydata.org/generated/seaborn.displot.html
        * ``catplot``:      seaborn.pydata.org/generated/seaborn.catplot.html
        * ``lmplot``:       seaborn.pydata.org/generated/seaborn.lmplot.html

    Other plots:
        * ``clustermap``:   seaborn.pydata.org/generated/seaborn.clustermap.html
        * ``pairplot``:     seaborn.pydata.org/generated/seaborn.pairplot.html
        * ``jointplot``:    seaborn.pydata.org/generated/seaborn.jointplot.html

    Args:
        data (dict): The data transformation framework results, expecting a
            single entry ``data`` which can be a pandas.DataFrame or an
            xarray data type.
        hlpr (PlotHelper): The plot helper instance
        sns_kind (str): Which seaborn plot to use
        free_indices (Tuple[str]): Which index names *not* to associate with a
            layout encoding; seaborn uses these to calculate the distribution
            statistics.
        optional_free_indices (Tuple[str], optional): These indices will be
            added to the free indices *if they are part of the data frame*.
            Otherwise, they are silently ignored.
        auto_encoding (Union[bool, dict], optional): Auto-encoding options.
        reset_index (bool, optional): Whether to reset indices such
            that only the ``free_indices`` remain as indices and all others are
            converted into columns.
        to_dataframe_kwargs (dict, optional): For xarray data types, this is
            used to convert the given data into a pandas.DataFrame.
        sample (bool, optional): If True, will sample a subset from the final
            dataframe, controlled by ``sample_kwargs``
        sample_kwargs (dict, optional): Passed to ``pd.DataFrame.sample``.
        **plot_kwargs: Passed on to the selected plotting function.
    """
    df = data["data"]

    # For xarray types, attempt conversion
    if isinstance(df, (xr.Dataset, xr.DataArray)):
        tdf_kwargs = to_dataframe_kwargs if to_dataframe_kwargs else {}
        log.note("Attempting conversion to pd.DataFrame ...")
        log.remark(
            "  Arguments:  %s",
            ", ".join(f"{k}: {v}" for k, v in tdf_kwargs.items()),
        )
        df = df.to_dataframe(**tdf_kwargs)

    # Re-index to get long-form data
    # See:  https://seaborn.pydata.org/tutorial/data_structure.html
    log.note("Evaluating data frame ...")
    log.remark("  Length:           %d", len(df))
    log.remark("  Shape:            %s", df.shape)
    log.remark("  Size:             %d", df.size)
    try:
        log.remark("  Columns:          %s", ", ".join(df.columns))
    except:  # TODO Make more specific or even avoid try-except
        log.remark("  Columns:          (none)")

    try:
        log.remark("  Indices:          %s", ", ".join(df.index.names))
    except:  # TODO Make more specific or even avoid try-except
        log.remark("  Indices:          (no named indices)")

    log.remark("  Free indices:     %s", ", ".join(free_indices))
    log.remark("  Optionally free:  %s", ", ".join(optional_free_indices))

    # TODO Add an option to make all indices free, excluding some ...

    # Apply optionally free indices
    free_indices += [n for n in optional_free_indices if n in df.index.names]

    # For some kinds, it makes sense to re-index such that only the free
    # indices are used as columns
    if reset_index:
        reset_for = [n for n in df.index.names if n not in free_indices]
        if reset_for:
            df = df.reset_index(level=reset_for)
            log.remark("  Reset index for:  %s", ", ".join(reset_for))

    # Might want to drop null values
    if dropna:
        dropna_kwargs = dropna_kwargs if dropna_kwargs else {}
        log.note("Dropping null values ...")
        log.remark(
            "  Arguments:  %s",
            ", ".join(f"{k}: {v}" for k, v in dropna_kwargs.items()),
        )
        df = df.dropna(**dropna_kwargs)
        log.remark("  Length after drop:  %d", len(df))

    # Sampling
    if sample:
        if not sample_kwargs:
            sample_kwargs = {}
            if isinstance(sample, int) and sample < len(df):
                sample_kwargs["n"] = sample

        if sample_kwargs:
            log.note("Sampling from data frame ...")
            log.remark(
                "  Arguments:  %s",
                ", ".join(f"{k}: {v}" for k, v in sample_kwargs.items()),
            )
            len_before = len(df)
            try:
                df = df.sample(**sample_kwargs)
            except Exception as exc:
                log.error(
                    "  Sampling failed with %s: %s", type(exc).__name__, exc
                )
            else:
                log.remark(
                    "  Sampling succeeded. New length: %d (%d)",
                    len(df),
                    len(df) - len_before,
                )
        else:
            log.note("Sampling skipped (no arguments applicable).")

    # ... further preprocessing ...

    # Interface with auto-encoding
    # Need to pop any given `kind` argument (valid input to sns.pairplot)
    kind = plot_kwargs.pop("kind", None)
    plot_kwargs = determine_encoding(
        {
            n: s
            for n, s in zip(
                df.index.names, getattr(df.index, "levshape", [len(df.index)])
            )
            if n not in free_indices
        },
        kind=sns_kind,
        auto_encoding=auto_encoding,
        default_encodings=SNS_ENCODINGS,
        plot_kwargs=plot_kwargs,
    )
    if kind is not None:
        plot_kwargs["kind"] = kind

    # Depending on plot kinds, determine some further arguments
    if kind in SNS_FACETGRID_KINDS:
        # Provide a best guess for the `x` encoding, if it is not given
        if "x" not in plot_kwargs and len(df.columns) == 1:
            x = str(df.columns[0])
            log.note("Using '%s' for x-axis encoding.", x)
            plot_kwargs["x"] = x

    # Retrieve the plot function
    try:
        plot_func = SNS_PLOT_FUNCS[sns_kind]

    except KeyError:
        _avail = ", ".join(SNS_PLOT_FUNCS)
        raise ValueError(
            f"Invalid plot kind '{sns_kind}'! Available: {_avail}"
        )

    # Actual plotting . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
    # Close the existing figure; the seaborn functions create their own
    hlpr.close_figure()

    # Let seaborn do the plotting
    log.note("Now invoking sns.%s ...", sns_kind)

    try:
        with figure_leak_prevention():
            fg = plot_func(data=df, **plot_kwargs)

    except Exception as exc:
        raise PlottingError(
            f"sns.{sns_kind} failed! Got {type(exc).__name__}: {exc}\n\n"
            f"Data was:\n{df}\n\n"
            f"Plot function arguments were:\n  {plot_kwargs}"
        ) from exc

    # Attach the created figure, including a workaround for `col_wrap`, in
    # which case `fg.axes` is one-dimensional (for whatever reason)
    if isinstance(fg, sns.JointGrid):
        fig = fg.fig
        axes = [[fg.ax_joint]]  # TODO consider registering all axes

    else:
        # Assume it's FacetGrid-like
        fig = fg.fig
        axes = fg.axes
        if axes.ndim != 2:
            axes = axes.reshape((fg._nrow, fg._ncol))

    hlpr.attach_figure_and_axes(fig=fig, axes=axes)

    # TODO Animation?!



"""This module provides matplotlib-related helper constructs"""

import copy
import logging
from typing import Union

import matplotlib as mpl
import matplotlib.patches as mpatches
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.legend_handler import HandlerPatch

log = logging.getLogger(__name__)

# -----------------------------------------------------------------------------

NORMS = {
    "Normalize": mpl.colors.Normalize,
    "BoundaryNorm": mpl.colors.BoundaryNorm,
    "NoNorm": mpl.colors.NoNorm,
    "LogNorm": mpl.colors.LogNorm,
    "PowerNorm": mpl.colors.PowerNorm,
    "SymLogNorm": mpl.colors.SymLogNorm,
    "TwoSlopeNorm": mpl.colors.TwoSlopeNorm,
}
"""matplotlib color normalizations supported by the :py:class:`.ColorManager`.
"""

# -----------------------------------------------------------------------------


class HandlerEllipse(HandlerPatch):
    """Custom legend handler to turn an ellipse handle into a legend key."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def create_artists(
        self,
        legend,
        orig_handle,
        xdescent,
        ydescent,
        width,
        height,
        fontsize,
        trans,
    ):
        """Create an ellipse as a matplotlib artist object."""
        center = 0.5 * width - 0.5 * xdescent, 0.5 * height - 0.5 * ydescent
        p = mpatches.Ellipse(
            xy=center, width=height + xdescent, height=height + ydescent
        )
        self.update_prop(p, orig_handle, legend)
        p.set_transform(trans)
        return [p]


# -----------------------------------------------------------------------------


class ColorManager:
    """Custom color manager which provides an interface to the
    :py:mod:`matplotlib.colors` module and aims to simplify working with
    colormaps, colorbars, and different normalizations.
    """

    def __init__(
        self,
        *,
        cmap: Union[str, dict, mpl.colors.Colormap] = None,
        norm: Union[str, dict, mpl.colors.Normalize] = None,
        labels: dict = None,
        vmin: float = None,
        vmax: float = None,
    ):
        """Initializes a :py:class:`.ColorManager` by building the colormap,
        the norm, and the colorbar labels.

        Args:
            cmap (Union[str, dict, matplotlib.colors.Colormap], optional):
                The colormap. If it is a string, it must name a registered
                colormap. If it is a dict, the following arguments are
                available:

                name (str, optional):
                    Name of a registered colormap.
                from_values (Union[dict, list], optional):
                    Dict of colors keyed by bin-specifier. If given, ``name``
                    is ignored and a discrete colormap is created from the list
                    of specified colors. The ``norm`` is then set to
                    :py:class:`matplotlib.colors.BoundaryNorm`.

                    The bins can be specified either by bin-centers (Scalar) or
                    by bin-intervals (2-tuples). For the former, the deduced
                    bin-edges are assumed halfway between the bin-centers. For
                    the latter, the given intervals must be pairwise connected.
                    In both cases, the bins must monotonically increase.

                    If a list of colors is passed they are automatically
                    assigned to the bin-centers ``[0, 1, 2, ...]``.
                under (Union[str, dict], optional): Passed on to
                    :py:meth:`matplotlib.colors.Colormap.set_under`
                over (Union[str, dict], optional): Passed on to
                    :py:meth:`matplotlib.colors.Colormap.set_over`
                bad (Union[str, dict], optional): Passed on to
                    :py:meth:`matplotlib.colors.Colormap.set_bad`
                placeholder_color (optional): ``None`` values in
                    ``from_values`` are replaced with this color
                    (default: white).

            norm (Union[str, dict, matplotlib.colors.Normalize], optional):
                The norm that is applied for the color-mapping. If it is a
                string, the matching norm in :py:mod:`matplotlib.colors`
                is created with default values.
                If it is a dict, the ``name`` entry specifies the norm and all
                further entries are passed to its constructor.
                Overwritten if a discrete colormap is specified via
                ``cmap.from_values``.
            labels (Union[dict, list], optional): Colorbar tick-labels keyed by
                tick position. If a list of labels is passed they are
                automatically assigned to the positions [0, 1, 2, ...].
            vmin (float, optional): The lower bound of the color-mapping.
                Ignored if norm is *BoundaryNorm*.
            vmax (float, optional): The upper bound of the color-mapping.
                Ignored if norm is *BoundaryNorm*.
        """
        cmap_kwargs = {}
        norm_kwargs = {}

        # If Colormap instance is given, set it directly. Otherwise parse the
        # cmap_kwargs below.
        if isinstance(cmap, mpl.colors.Colormap):
            self._cmap = cmap

        elif isinstance(cmap, str) or cmap is None:
            cmap_kwargs = dict(name=cmap)

        else:
            cmap_kwargs = copy.deepcopy(cmap)

        # If Normalize instance is given, set it directly. Otherwise parse the
        # norm_kwargs below.
        if isinstance(norm, mpl.colors.Normalize):
            self._norm = norm

        elif isinstance(norm, str) or norm is None:
            norm_kwargs = dict(name=norm)

        else:
            norm_kwargs = copy.deepcopy(norm)

        # Parse configuration for custom discrete colormapping
        if "from_values" in cmap_kwargs:
            mapping = cmap_kwargs.pop("from_values")

            # Parse shortcut notation
            if isinstance(mapping, list):
                mapping = {k: v for k, v in enumerate(mapping)}

            # Replace all None entries by the placeholder color. If not given,
            # set white as default.
            _placeholder_color = cmap_kwargs.pop("placeholder_color", "w")
            mapping = {
                k: (v if v is not None else _placeholder_color)
                for k, v in mapping.items()
            }

            cmap_kwargs["name"] = "ListedColormap"
            cmap_kwargs["colors"] = list(mapping.values())

            # Overwrite the norm configuration
            norm_kwargs = {
                "name": "BoundaryNorm",
                "ncolors": len(mapping),
                "boundaries": self._parse_boundaries(list(mapping.keys())),
            }

            log.remark(
                "Configuring a discrete colormap 'from values'. "
                "Overwriting 'norm' to BoundaryNorm with %d colors.",
                norm_kwargs["ncolors"],
            )

        # BoundaryNorm has no vmin/vmax argument
        if not norm_kwargs.get("name", None) == "BoundaryNorm":
            norm_kwargs["vmin"] = vmin
            norm_kwargs["vmax"] = vmax

            log.remark("norm.vmin and norm.vmax set to %s and %s.", vmin, vmax)

        # Parse shortcut notation
        if isinstance(labels, list):
            labels = {k: v for k, v in enumerate(labels)}

        self._labels = copy.deepcopy(labels)

        # Set cmap and norm if not done already
        if cmap_kwargs:
            self._cmap = self._create_cmap(**cmap_kwargs)

        if norm_kwargs:
            self._norm = self._create_norm(**norm_kwargs)

    @property
    def cmap(self):
        return self._cmap

    @property
    def norm(self):
        return self._norm

    @property
    def labels(self):
        return self._labels

    def _parse_boundaries(self, bins):
        """Parses the boundaries for the BoundaryNorm.

        Args:
            bins (Sequence): Either monotonically increasing sequence of bin
                centers or sequence of connected intervals (2-tuples).

        Returns:
            (list): Monotonically increasing boundaries.

        Raises:
            ValueError: On disconnected intervals or decreasing boundaries.
        """

        def from_intervals(intervals):
            """Extracts bin edges from sequence of connected intervals"""
            b = [intervals[0][0]]

            for low, up in intervals:
                if up < low:
                    raise ValueError(
                        "Received decreasing boundaries: " f"{up} < {low}."
                    )

                elif b[-1] != low:
                    raise ValueError(
                        "Received disconnected intervals: Upper "
                        f"bound {b[-1]} and lower bound {low} of "
                        "the proximate interval do not match."
                    )

                b.append(up)

            return b

        def from_centers(centers):
            """Calculates the bin edges as the halfway points between adjacent
            bin centers."""
            centers = np.array(centers)

            if len(centers) < 2:
                raise ValueError(
                    "At least 2 bin centers must be given to "
                    f"create a BoundaryNorm. Got: {centers}"
                )

            halves = 0.5 * np.diff(centers)

            b = (
                [centers[0] - halves[0]]
                + [c + h for c, h in zip(centers, halves)]
                + [centers[-1] + halves[-1]]
            )

            return b

        if isinstance(bins[0], tuple):
            boundaries = from_intervals(bins)
        else:
            boundaries = from_centers(bins)

        return boundaries

    def _create_cmap(
        self,
        name: str = None,
        *,
        bad: Union[str, dict] = None,
        under: Union[str, dict] = None,
        over: Union[str, dict] = None,
        **cmap_kwargs,
    ):
        """Creates a colormap.

        Args:
            name (str, optional): The colormap name. Can either be the name of
                a registered colormap or ``ListedColormap``. ``None`` means
                that the ``image.cmap`` value from the RC parameters is used.
            bad (Union[str, dict], optional): Set color to be used for masked
                values.
            under (Union[str, dict], optional): Set the color for low
                out-of-range values when ``norm.clip = False``.
            over (Union[str, dict], optional): Set the color for high
                out-of-range values when ``norm.clip = False``.
            **cmap_kwargs: If ``name = ListedColormap``, passed on to the
                constructor of the colormap, else passed to
                matplotlib.cm.get_cmap.

        Returns:
            matplotlib.colors.Colormap: The created colormap.

        Raises:
            ValueError: On invalid colormap name.
        """
        if name == "ListedColormap":
            cmap = mpl.colors.ListedColormap(**cmap_kwargs)

        else:
            try:
                cmap = copy.copy(mpl.cm.get_cmap(name, **cmap_kwargs))

            except ValueError as err:
                raise ValueError(
                    f"Received invalid colormap name: '{name}'. "
                    "Must name a registered colormap."
                ) from err

        if isinstance(bad, str):
            bad = dict(color=bad)

        if isinstance(under, str):
            under = dict(color=under)

        if isinstance(over, str):
            over = dict(color=over)

        if bad is not None:
            cmap.set_bad(**bad)

        if under is not None:
            cmap.set_under(**under)

        if over is not None:
            cmap.set_over(**over)

        return cmap

    def _create_norm(self, name: str = None, **norm_kwargs):
        r"""Creates a norm.

        Args:
            name (str, optional): The norm name. Must name a
                matplotlib.colors.Normalize instance (see
                `matplotlib.colors <https://matplotlib.org/api/colors_api.html>`_).
                ``None`` means *Normalize*.
            \**norm_kwargs: Passed on to the constructor of the norm.

        Returns:
            matplotlib.colors.Normalize: The created norm.

        Raises:
            ValueError: On invalid norm specification.
        """
        if name is None:
            name = "Normalize"

        if name not in NORMS:
            available_norms = ", ".join(NORMS.keys())
            raise ValueError(
                f"Received invalid norm specifier: '{name}'. "
                f"Must be one of: {available_norms}"
            )

        return NORMS[name](**norm_kwargs)

    def map_to_color(self, X: Union[float, np.ndarray]):
        """Maps the input data to color(s) by applying both norm and colormap.

        Args:
            X (Union[float, numpy.ndarray]): Data value(s) to convert to RGBA.

        Returns:
            Tuple of RGBA values if X is scalar, otherwise an array of RGBA
            values with a shape of ``X.shape + (4, )``.
        """
        return self.cmap(self.norm(X))

    def create_cbar(
        self,
        mappable,
        *,
        fig=None,
        ax=None,
        label: str = None,
        label_kwargs: dict = None,
        tick_params: dict = None,
        **cbar_kwargs,
    ):
        """Creates a colorbar.

        Args:
            mappable: The matplotlib.cm.ScalarMappable described by the
                colorbar.
            fig (None, optional): The figure
            ax (None, optional): The axis
            label (str, optional): A label for the colorbar
            label_kwargs (dict, optional): Additional parameters passed to
                :py:meth:`matplotlib.colorbar.Colorbar.set_label`
            tick_params (dict, optional): Set colorbar tick parameters via
                `cb.ax.tick_params <https://matplotlib.org/stable/api/_as_gen/matplotlib.axes.Axes.tick_params.html>`__
            **cbar_kwargs: Passed on to :py:func:`matplotlib.pyplot.colorbar`

        Returns:
            The created :py:class:`matplotlib.colorbar.Colorbar` object
        """
        if fig is None:
            fig = plt.gcf()

        if ax is None:
            ax = fig.gca()

        cb = fig.colorbar(mappable, ax=ax, **cbar_kwargs)

        # Colorbar label
        if label:
            cb.set_label(label=label, **(label_kwargs if label_kwargs else {}))

        # Ticks and tick labels
        if self.labels is not None:
            cb.set_ticks(list(self.labels.keys()))
            cb.set_ticklabels(list(self.labels.values()))

        if tick_params:
            cb.ax.tick_params(**tick_params)

        return cb



"""This module provides plotting functions to visualize cellular automata."""

import copy
import logging
from math import ceil, sqrt
from typing import Callable, Dict, Sequence, Union

import matplotlib as mpl
import numpy as np
import xarray as xr
from dantro.abc import AbstractDataContainer
from matplotlib.collections import RegularPolyCollection
from matplotlib.colors import ListedColormap
from mpl_toolkits.axes_grid1 import make_axes_locatable

from ...tools import recursive_update
from .. import DataManager, UniverseGroup
from . import PlotHelper, UniversePlotCreator, is_plot_func

log = logging.getLogger(__name__)

# Increase log threshold for animation module
logging.getLogger("matplotlib.animation").setLevel(logging.WARNING)


# -----------------------------------------------------------------------------


def _get_ax_size(ax, fig) -> tuple:
    """The width and height of the given axis in pixels"""
    bbox = ax.get_window_extent().transformed(fig.dpi_scale_trans.inverted())
    width, height = bbox.width, bbox.height
    width *= fig.dpi
    height *= fig.dpi
    return width, height


def imshow_hexagonal(
    data: xr.DataArray,
    *,
    hlpr: PlotHelper,
    colormap: Union[str, mpl.colors.Colormap],
    **kwargs,
) -> mpl.image.AxesImage:
    """Display data as an image, i.e., on a 2D hexagonal grid.

    Args:
        data (xarray.DataArray): The array-like data to plot as image.
        hlpr (PlotHelper): The plot helper.
        colormap (str or matplotlib.colors.Colormap): The colormap to use.

    Returns:
        The RegularPolyCollection representing the hexgrid.
    """
    width, height = _get_ax_size(hlpr.ax, hlpr.fig)
    s = (height / data.y.size) / 0.75 / 2
    # NOTE the 0.75 factor is required because of the hexagonal offset geometry
    area = 3**1.5 / 2 * s**2

    # distinguish pair and impair rows (impair have offset)
    hex_s = 2 * data.isel(y=0).y
    y_ids = ((data.y / hex_s - 0.5) / 0.75).round().astype(int)

    # compute offsets of polygons
    xx, yy = np.meshgrid(data.x, data.y)
    x_offsets = xr.DataArray(
        data=xx, dims=("y", "x"), coords=dict(x=data.x, y=data.y)
    )
    y_offsets = xr.DataArray(
        data=yy, dims=("y", "x"), coords=dict(x=data.x, y=data.y)
    )

    # ... and add an x-offset for impair rows
    x_origin = data.isel(x=0, y=0).coords["x"]
    x_offsets[y_ids % 2 == 1] += x_origin

    # # assign the true coordinates
    # d = d.assign_coords(x=('x', d.x))
    # d_offset = d_offset.assign_coords(x=('x', d_offset.x + x_origin))

    # get the color mapping
    if isinstance(colormap, str):
        cmap = mpl.cm.get_cmap(name=colormap)
    else:
        cmap = colormap
    map_color = mpl.cm.ScalarMappable(cmap=cmap)

    pcoll = RegularPolyCollection(
        6,
        sizes=(area,),
        rotation=0,
        facecolor=map_color.to_rgba(data.data.flatten()),
        offsets=np.transpose(
            [x_offsets.data.flatten(), y_offsets.data.flatten()]
        ),
        transOffset=hlpr.ax.transData,
        animated=kwargs.get("animated", True),
        rasterized=kwargs.get("rasterized", False),
    )
    hlpr.ax.add_collection(pcoll)

    # use same length scale in x and y
    hlpr.ax.set_aspect("equal")

    # rescale cmap
    im = mpl.image.AxesImage(hlpr.ax)

    im.set_cmap(colormap)
    if "vmin" in kwargs and "vmax" in kwargs:
        # From `limits` argument; will either have none or both
        im.set_clim(kwargs["vmin"], kwargs["vmax"])

    return im


def _plot_ca_property(
    prop_name: str,
    *,
    hlpr: PlotHelper,
    data: xr.DataArray,
    default_imshow_kwargs: dict,
    cax=None,
    grid_structure: str = None,
    limits: list = None,
    cmap: Union[str, dict] = "viridis",
    draw_cbar: bool = True,
    title: str = None,
    no_cbar_markings: bool = False,
    imshow_kwargs: dict = None,
    **cbar_kwargs,
):
    """Helper function to plot a property on a given axis and return
    an imshow object

    TODO Finish writing docstring

    Args:
        prop_name (str): The property to plot
        hlpr (PlotHelper): Description
        data (xarray.DataArray): The array-like data to plot as image
        grid_structure (str, optional): Description
        default_imshow_kwargs (dict): Description
        cax: colorbar axis object
        limits (list, optional): The imshow limits to use; will also be
            the limits of the colorbar.
        cmap (Union[str, dict], optional): The colormap to use. If a dict is
            given, defines a (discrete) ``ListedColormap`` from the values.
        draw_cbar (bool, optional): whether to draw a color bar
        title (str, optional): The subplot figure title
        no_cbar_markings (bool, optional): Whether to suppress colorbar
            markings (ticks and tick labels)
        imshow_kwargs (dict, optional): Passed to ``plt.imshow``
        **cbar_kwargs: Passed to ``fig.colorbar``

    Returns:
        imshow object

    Raises:
        TypeError: For invalid ``cmap`` argument.
        ValueError: For invalid grid structure; needs to be ``square``,
            ``hexagonal`` or ``triangular``.
    """
    # Get colormap, either a continuous or a discrete one
    if isinstance(cmap, str):
        norm = None
        bounds = None
        colormap = cmap

    elif isinstance(cmap, dict):
        colormap = ListedColormap(cmap.values())
        bounds = limits
        norm = mpl.colors.BoundaryNorm(bounds, colormap.N)

    else:
        raise TypeError(
            "Argument cmap needs to be either a string with name of the "
            "colormap or a dict with values for a discrete colormap! "
            f"Was: {type(cmap)} with value: '{cmap}'"
        )

    # Parse additional colorbar kwargs and set some default values
    add_cbar_kwargs = dict()
    if "fraction" not in cbar_kwargs:
        add_cbar_kwargs["fraction"] = 0.05

    if "pad" not in cbar_kwargs:
        add_cbar_kwargs["pad"] = 0.02

    # Fill imshow_kwargs, using defaults
    imshow_kwargs = imshow_kwargs if imshow_kwargs else {}
    imshow_kwargs = recursive_update(
        copy.deepcopy(imshow_kwargs),
        default_imshow_kwargs if default_imshow_kwargs else {},
    )
    if limits:
        vmin, vmax = limits
        imshow_kwargs["vmin"] = vmin
        imshow_kwargs["vmax"] = vmax

    # Create imshow object on the currently selected axis
    grid_structure = (
        grid_structure
        if grid_structure
        else data.attrs.get("grid_structure", "square")
    )
    if grid_structure == "square" or grid_structure is None:
        im = hlpr.ax.imshow(
            data.T,
            cmap=colormap,
            animated=True,
            rasterized=True,
            origin="lower",
            aspect="equal",
            **imshow_kwargs,
        )

    elif grid_structure == "hexagonal":
        hlpr.ax.clear()
        im = imshow_hexagonal(
            data=data,
            hlpr=hlpr,
            animated=True,
            rasterized=True,
            colormap=colormap,
            **imshow_kwargs,
        )
        title = title if title else prop_name
        hlpr.ax.set_title(title)

    elif grid_structure == "triangular":
        raise ValueError("Plotting of triangular grid not implemented!")

    else:
        raise ValueError(f"Unknown grid structure '{grid_structure}'!")

    # Create the colorbar.
    # For hexagonal grids, manually create a separate axis next to the
    # current plotting axis. Add a cbar and manually set the ticks and
    # boundaries (as the cbar returned by the mpl.image.AxesImage has
    # default range (0, 1))
    if draw_cbar and grid_structure == "hexagonal":
        cax.clear()
        if bounds:
            num_colors = len(cmap)
            boundaries = [i / num_colors for i in range(num_colors + 1)]
            tick_locs = [
                (2 * i + 1) / (2 * num_colors) for i in range(num_colors)
            ]
        else:
            boundaries = None
            tick_locs = [i * 0.25 for i in range(5)]
            if isinstance(colormap, str):
                colormap = mpl.cm.get_cmap(name=colormap)

        cbar = mpl.colorbar.ColorbarBase(
            cax, cmap=colormap, boundaries=boundaries
        )
        cbar.set_ticks(tick_locs)

        # For discrete colorbars, set the tick labels at the positions
        # defined; for continuous colorbars, get the upper and lower
        # boundaries from the dataset.
        if bounds:
            cbar.ax.set_yticklabels(cmap.keys())
        else:
            lower = np.min(data.data)
            upper = np.max(data.data)
            if lower == upper:
                diff = 0.01
            else:
                diff = upper - lower

            ticklabels = [i * diff / 4 + lower for i in range(5)]
            cbar.ax.set_yticklabels(ticklabels)

    elif draw_cbar:
        cbar = hlpr.fig.colorbar(
            im, ax=hlpr.ax, ticks=bounds, **cbar_kwargs, **add_cbar_kwargs
        )
        # For a discrete colormap, adjust the tick positions
        if bounds:
            num_colors = len(cmap)
            tick_locs = (
                (np.arange(num_colors) + 0.5) * (num_colors - 1) / num_colors
            )
            cbar.set_ticks(tick_locs)
            cbar.ax.set_yticklabels(cmap.keys())

    # Remove markings, if configured to do so
    if draw_cbar and no_cbar_markings:
        cbar.set_ticks([])
        cbar.ax.set_yticklabels([])

    # Remove main axis labels and ticks
    hlpr.ax.axis("off")

    # Provide configuration options to plot helper
    hlpr.provide_defaults("set_title", title=(title if title else prop_name))

    return im


# -----------------------------------------------------------------------------


@is_plot_func(use_dag=True, supports_animation=True)
def caplot(
    *,
    hlpr: PlotHelper,
    data: dict,
    to_plot: Dict[str, dict],
    from_dataset: xr.Dataset = None,
    frames: str = "time",
    frames_isel: Union[int, Sequence] = None,
    grid_structure: str = None,
    aspect: float = 1.0,
    aspect_pad: float = 0.1,
    size: float = None,
    col_wrap: Union[int, str, bool] = "auto",
    default_imshow_kwargs: dict = None,
    default_cbar_kwargs: dict = None,  # TODO implement
    suptitle_fstr: str = "{} = {}",
    suptitle_kwargs: dict = None,
):
    """Plots an animated series of one or many 2D Cellular Automata states.

    The data used for plotting is assembled from ``data`` using the keys that
    are specified in ``to_plot``. Alternatively, the ``from_dataset`` argument
    can be used to pass a dataset which contains all the required data.

    The keys in ``to_plot`` should match the names of data variables.
    The values in ``to_plot`` specify the individual subplots' properties like
    the color map that is to be used or the minimum or maximum values.

    For plotting, the matplotlib ``imshow`` function is used.

    .. note::

        Overall, there should be two spatial dataset dimensions and one that
        goes along the ``frames`` dimension. All coordinate labels should be
        identical, otherwise the behavior is not defined.

    Args:
        hlpr (PlotHelper): The plot helper instance
        data (dict): The selected data
        to_plot (Dict[str, dict]): Which data to plot and how. The keys of
            this dict refer to an item within the selected ``data`` or the
            given dataset.
            Each of these keys is expected to hold yet another dict,
            supporting the following configuration options (all optional):

                - ``cmap`` (str or dict): The colormap to use. If it is a
                    dict, a discrete colormap is assumed. The keys will be the
                    labels and the values the color. Association happens in
                    the order of entries.
                - ``title`` (str): The title for this sub-plot
                - ``limits`` (2-tuple, list): The fixed heat map limits of this
                    property; if not given, limits will be auto-scaled. If
                    they are ``min`` or ``max``, the *global* minimum or
                    maximum, respectively, will be used.
                    Note that specifying ``limits`` will overwrite potentially
                    existing ``vmin`` and ``vmax`` arguments to imshow.
                - ``**imshow_kwargs``: passed on to imshow invocation

        from_dataset (xarray.Dataset, optional): If given, will use this object
            instead of assembling a dataset from ``data`` and ``to_plot`` keys.
        frames (str, optional): Name of the animated dimension, typically the
            time dimension.
        frames_isel (Union[int, Sequence], optional): The index selector for
            the frames dimension. Can be a single integer but also a range
            expression.
        grid_structure (str, optional): The underlying grid structure, can be
            ``square``, ``hexagonal``, or ``triangular`` (not implemented).
            If None, will try to read it from the individual properties' data
            attribute ``grid_structure``.
        aspect (float, optional): The aspect ratio (width/height) of the
            subplots; should match the aspect ratio of the data.
        aspect_pad (float, optional): A factor that is added to the calcuation
            of the subplots width. This can be used to create a horizontal
            padding between subplots.
        size (float, optional): The height in inches of a subplot.
            Is used to determine the subplot size with the width being
            calculated by ``size * (aspect + aspect_pad)``.
        col_wrap (Union[int, str, bool], optional): Controls column wrapping.
            If ``auto``, will compute a column wrapping that leads to an
            (approximately) square subplots layout (not taking into account
            the subplots aspect ratio, only the grid layout). This will start
            producing a column wrapping with four or more properties to plot.
        default_imshow_kwargs (dict, optional): The default parameters passed
            to the underlying imshow plotting function. These are updated by
            the values given via ``to_plot``.
        default_cbar_kwargs (dict, optional): NOT IMPLEMENTED  # FIXME
        suptitle_fstr (str, optional): A format string used to create the
            suptitle string. Passed arguments are ``frames`` and the currently
            selected frames *coordinate* (not the index!).
            If this evaluates to False, will not create a figure suptitle.
        suptitle_kwargs (dict, optional): Passed on to ``fig.suptitle``.
    """
    # Helper functions ........................................................

    def get_grid_structure(d: xr.DataArray) -> str:
        """Retrieves the grid structure from data attributes"""
        grid_structure = d.attrs.get("grid_structure")
        if isinstance(grid_structure, np.ndarray):
            grid_structure = grid_structure.item()

        return grid_structure

    def prepare_data(data: dict, *, prop_name: str) -> xr.DataArray:
        """Prepares data for (later) creating an :py:class:`xarray.Dataset`"""
        d = data[prop_name]
        if isinstance(d, AbstractDataContainer):
            d = d.data
        return d

    def select_data(ds: xr.Dataset, name: str, isel: dict) -> xr.DataArray:
        """Selects a slice of data for plotting using
        :py:meth:`~xarray.DataArray.isel` on the data variable ``name``."""
        return ds[name].isel(isel)

    def set_suptitle(data: xr.DataArray):
        """Sets the suptitle"""
        if not suptitle_fstr:
            return

        hlpr.fig.suptitle(
            suptitle_fstr.format(frames, data.coords[frames].item()),
            **(suptitle_kwargs if suptitle_kwargs else {}),
        )

    # Prepare the data ........................................................
    # Work on a copy of the plot spec
    to_plot = copy.deepcopy(to_plot)

    # Bring data into xr.Dataset form
    log.note("Preparing data for CA plot ...")
    if from_dataset:
        log.remark("Using explicitly passed dataset ...")
        ds = from_dataset
    else:
        log.remark("Constructing dataset ...")
        ds = xr.Dataset({p: prepare_data(data, prop_name=p) for p in to_plot})

    # Check that frames dimension is available
    if not frames or frames not in ds.coords:
        _avail = ", ".join(xr.coords.keys())
        raise ValueError(
            f"Invalid frames coordinate dimension '{frames}'! "
            f"Available coordinates:  {_avail}"
        )

    # Apply selection along frames dimension
    if frames_isel is not None:
        if isinstance(frames_isel, int):
            frames_isel = [frames_isel]
        _selector = {frames: frames_isel}
        log.remark("Applying index selection %s  ...", _selector)
        ds = ds.isel(_selector, drop=False)

    # TODO x-y mapping?
    # TODO Automate aspect computation from x and y dimensions

    # Depending on length of coordinate dimension, ensure that animation mode
    # is enabled or disabled.
    num_frames = ds.coords[frames].size
    if num_frames > 1:
        hlpr.enable_animation()
    else:
        hlpr.disable_animation()

    # If not given, retrieve the structure from the data variable's attributes.
    if not grid_structure:
        structures = {p: get_grid_structure(ds[p]) for p in to_plot}

        if len(set(structures.values())) != 1:
            raise ValueError(
                "Mismatch in grid structure; all grid structures need to be "
                f"the same but were:  {structures}"
            )
        grid_structure = next(iter(structures.values()))

    # Evaluate limits argument for all properties
    for prop_name, spec in to_plot.items():
        if spec.get("limits"):
            vmin, vmax = spec["limits"]
            if vmin == "min":
                vmin = ds[prop_name].min().item()
            if vmax == "max":
                vmax = ds[prop_name].max().item()
            spec["limits"] = (vmin, vmax)

    # Inform about the data that is to be plotted
    log.note(
        "Performing CA plot for %d data variable%s ...",
        len(to_plot),
        "" if len(to_plot) == 1 else "s",
    )
    log.remark("  Data variables:     %s", ", ".join(to_plot))
    log.remark(
        "  Dimensions:         %s",
        ", ".join(f"{k}: {s}" for k, s in ds.sizes.items()),
    )
    log.remark("  Grid structure:     %s", grid_structure)

    # Some final checks ...
    if len(ds.sizes) != 3:
        raise ValueError(
            f"Dataset shape needs to be 3-dimensional, but was: {ds.sizes}"
        )

    # Prepare the figure ......................................................
    # Evaluate column wrapping
    if col_wrap and not (col_wrap == "auto" and len(to_plot) < 4):
        if col_wrap == "auto":
            col_wrap = ceil(sqrt(len(to_plot)))
        log.remark("  Column wrapping:    %s", col_wrap)

        ncols = col_wrap
        nrows = ceil(len(to_plot) / col_wrap)
        axis_map = {
            p: dict(col=i % col_wrap, row=i // col_wrap)
            for i, p in enumerate(to_plot)
        }

    else:
        ncols = len(to_plot)
        nrows = 1
        axis_map = {p: dict(col=col, row=0) for col, p in enumerate(to_plot)}

    # Determine the figsize from the size argument (height) and data aspect.
    # Depending on structure, may need to adapt figure size.
    if not size:
        size = mpl.rcParams["figure.figsize"][1]
    figsize = (size * (aspect + aspect_pad), size)

    if grid_structure == "hexagonal":
        figsize = (figsize[0] * 1.2, figsize[1])

    # Create the figure and set all axes as invisible. This is needed because
    # col_wrap may lead to some subplots being completely empty.
    # Below
    hlpr.setup_figure(
        figsize=figsize,
        scale_figsize_with_subplots_shape=True,
        ncols=ncols,
        nrows=nrows,
    )
    for ax in hlpr.axes.flat:
        ax.set_visible(False)

    # Do the single plot for all data variables, looping through subfigures.
    # This creates the first imshow objects, which are being kept track of
    # such that they can later be updated.
    # This also sets the axes as visible again, but only for those that have
    # a property assigned to them.
    ims = dict()
    for i, (prop_name, props) in enumerate(to_plot.items()):
        hlpr.select_axis(**axis_map[prop_name])
        hlpr.ax.set_visible(True)

        # For hexagonal grids, need to add custom colorbar axis here
        # TODO Parameterise default values
        if props.get("draw_cbar", True) and grid_structure == "hexagonal":
            divider = make_axes_locatable(hlpr.ax)
            props["cax"] = divider.append_axes("right", size="5%", pad=0.2)
            hlpr.select_axis(**axis_map[prop_name])

        # Select the appropriate data, then plot the data variable
        data = select_data(ds, prop_name, {frames: 0})

        ims[prop_name] = _plot_ca_property(
            prop_name,
            hlpr=hlpr,
            data=data,
            grid_structure=grid_structure,
            default_imshow_kwargs=default_imshow_kwargs,
            **props,
        )

        # Use this data for setting the figure suptitle
        if i == 0:
            set_suptitle(data)

    # End of single frame CA state plot function ..............................
    # The above variables are all available below, but the update function is
    # supposed to start plotting anew, starting from frame 0.

    def update_data():
        """Updates the data of the imshow objects"""
        log.note("Plotting animation with %d frames ...", num_frames)

        for frame_idx in range(num_frames):
            log.debug("Plotting frame %d ...", frame_idx)

            for i, (prop_name, props) in enumerate(to_plot.items()):
                hlpr.select_axis(**axis_map[prop_name])

                frame_data = select_data(ds, prop_name, {frames: frame_idx})

                # Depending on grid structure, update data or plot anew
                if grid_structure == "hexagonal":
                    # For hexagonal grids recreate the plot each time.
                    # Just resetting the data does not show the updated states
                    # otherwise because the facecolors have to be updated, too.
                    ims[prop_name] = _plot_ca_property(
                        prop_name,
                        data=frame_data,
                        hlpr=hlpr,
                        grid_structure=grid_structure,
                        default_imshow_kwargs=default_imshow_kwargs,
                        **props,
                    )

                else:
                    # Update imshow data without creating a new object
                    ims[prop_name].set_data(frame_data.T)

                    # If no limits are provided, autoscale the new limits in
                    # the case of continuous colormaps. A discrete colormap
                    # provided as a dict should never have to autoscale.
                    if not isinstance(props.get("cmap"), dict):
                        if not props.get("limits"):
                            ims[prop_name].autoscale()

                # Use the first subplot's data for setting the figure suptitle
                if i == 0:
                    set_suptitle(frame_data)

            # Done with this frame; yield control to the animation framework.
            yield

    # Register this update method with the helper, which takes care of the rest
    hlpr.register_animation_update(update_data)


# -----------------------------------------------------------------------------


@is_plot_func(creator_type=UniversePlotCreator, supports_animation=True)
def state(
    dm: DataManager,
    *,
    uni: UniverseGroup,
    hlpr: PlotHelper,
    model_name: str,
    to_plot: dict,
    time_idx: int,
    default_imshow_kwargs: dict = None,
    **_kwargs,
):
    r"""Plots the state of the cellular automaton as a 2D heat map.
    This plot function can be used for a single plot, but also supports
    animation.

    Which properties of the state to plot can be defined in ``to_plot``.

    Args:
        dm (DataManager): The DataManager that holds all loaded data
        uni (UniverseGroup): The currently selected universe, parsed by the
            :py:class:`~utopya.eval.plotcreators.UniversePlotCreator`.
        hlpr (PlotHelper): The plot helper
        model_name (str): The name of the model of which the data is to be
            plotted
        to_plot (dict): Which data to plot and how. The keys of this dict
            refer to a path within the data and can include forward slashes to
            navigate to data of submodels. Each of these keys is expected to
            hold yet another dict, supporting the following configuration
            options (all optional):

                - cmap (str or dict): The colormap to use. If it is a dict, a
                    discrete colormap is assumed. The keys will be the labels
                    and the values the color. Association happens in the order
                    of entries.
                - title (str): The title for this sub-plot
                - limits (2-tuple, list): The fixed heat map limits of this
                    property; if not given, limits will be auto-scaled.
                - \**imshow_kwargs: passed on to imshow invocation

        time_idx (int): Which time index to plot the data of. Is ignored when
            creating an animation.
        default_imshow_kwargs (dict, optional): The default parameters passed
            to the underlying imshow plotting function. These are updated by
            the values given via ``to_plot``.

    Raises:
        ValueError: Shape mismatch of data selected by ``to_plot``
        AttributeError: Got unsupported arguments (referring to the old data
            transformation framework)
    """
    if _kwargs:
        raise AttributeError(
            "This plot no longer supports preprocessing or transformation but "
            f"got one of the following arguments:  {_kwargs}\n"
            "Use the new CA plot function (.ca.caplot) which supports the "
            "data transformation framework."
        )

    log.warning(
        "The .ca.state plot is deprecated and should no longer be used!\n"
        "Please use the .ca.caplot function, which is almost identical in its "
        "interface and uses the data transformation framework for data "
        "selection and pre-processing."
    )

    # Helper functions ........................................................

    def prepare_data(
        prop_name: str, *, all_data: dict, time_idx: int
    ) -> np.ndarray:
        """Prepares the data for plotting"""
        return all_data[prop_name][time_idx]

    # Prepare the data ........................................................
    # Get the group that all datasets are in
    grp = uni["data"][model_name]

    # Collect all data
    all_data = {p: grp[p] for p in to_plot.keys()}
    shapes = [d.shape for p, d in all_data.items()]

    if any([shape != shapes[0] for shape in shapes]):
        raise ValueError(
            "Shape mismatch of properties {}: {}! Cannot plot."
            "".format(", ".join(to_plot.keys()), shapes)
        )

    # Can now be sure they all have the same shape,
    # so its fine to take the first shape to extract the number of steps
    num_steps = shapes[0][0]

    structure = prepare_data(
        list(to_plot.keys())[0], all_data=all_data, time_idx=0
    ).attrs.get("grid_structure", "square")

    # Prepare the figure ......................................................
    # Prepare the figure to have as many columns as there are properties
    hlpr.setup_figure(
        ncols=len(to_plot), scale_figsize_with_subplots_shape=True
    )
    if structure == "hexagonal":
        old_figsize = hlpr.fig.get_size_inches()  # (width, height)
        hlpr.fig.set_size_inches(
            old_figsize[0] * 1.25,
            old_figsize[1],
        )

    # Store the imshow objects such that only the data has to be updated in a
    # following iteration step. Keys will be the property names.
    ims = dict()

    # Do the single plot for all properties, looping through subfigures
    for col_no, (prop_name, props) in enumerate(to_plot.items()):
        # Select the axis
        hlpr.select_axis(col_no, 0)

        # For hexagonal grids, add custom colorbar axis
        if structure == "hexagonal":
            divider = make_axes_locatable(hlpr.ax)
            cax = divider.append_axes("right", size="5%", pad=0.2)
            hlpr.select_axis(col_no, 0)

        # Get the data for this time step
        data = prepare_data(prop_name, all_data=all_data, time_idx=time_idx)

        # In the first time step create a new imshow object
        ims[prop_name] = _plot_ca_property(
            prop_name,
            data=data,
            hlpr=hlpr,
            default_imshow_kwargs=default_imshow_kwargs,
            **props,
        )

    # End of single frame CA state plot function ..............................
    # The above variables are all available below, but the update function is
    # supposed to start plotting anew starting from frame 0.

    def update_data():
        """Updates the data of the imshow objects"""
        log.info(
            "Plotting animation with %d frames of %d %s each ...",
            num_steps,
            len(to_plot),
            "property" if len(to_plot) == 1 else "properties",
        )

        for time_idx in range(num_steps):
            log.debug("Plotting frame for time index %d ...", time_idx)

            # Loop through the columns
            for col_no, (prop_name, props) in enumerate(to_plot.items()):
                # Select the axis
                hlpr.select_axis(col_no, 0)

                # Get the data for this time step
                data = prepare_data(
                    prop_name, all_data=all_data, time_idx=time_idx
                )

                # Get the structure of the grid data
                structure = data.attrs.get("grid_structure", "square")

                # For hexagonal grids recreate the plot each time.
                # Just resetting the data does not show the updated states
                # otherwise because the facecolors have to be updated, too.
                # For other grid structures just update the data and colormap.
                if structure == "hexagonal":
                    ims[prop_name] = plot_property(
                        prop_name,
                        data=data,
                        hlpr=hlpr,
                        default_imshow_kwargs=default_imshow_kwargs,
                        **props,
                    )

                else:
                    # Update imshow data without creating a new object
                    ims[prop_name].set_data(data.T)

                    # If no limits are provided, autoscale the new limits in
                    # the case of continuous colormaps. A discrete colormap,
                    # that is provided as a dict, should never have to
                    # autoscale.
                    if not isinstance(props.get("cmap"), dict):
                        if not props.get("limits"):
                            ims[prop_name].autoscale()

            # Done with this frame; yield control to the animation framework
            # which will grab the frame...
            yield

        log.info("Animation finished.")

    # Register this update method with the helper, which takes care of the rest
    hlpr.register_animation_update(update_data)



"""Tests the utopya run CLI command"""

import logging
import time
import traceback

import pytest

from .. import ADVANCED_MODEL, DUMMY_MODEL
from .._fixtures import *
from . import invoke_cli


def _check_result(res, expected_exit: int = 0):
    if res.exception:
        print(res.output)
        traceback.print_tb(res.exception.__traceback__)

    elif res.exit_code != expected_exit:
        print(res.output)

    assert res.exit_code == expected_exit


# -----------------------------------------------------------------------------


def test_run(with_test_models, tmp_output_dir):
    """Tests the invocation of the utopya run command"""
    # Simplest case with the dummy model
    res = invoke_cli(("run", DUMMY_MODEL, "-d"))
    _check_result(res, expected_exit=0)
    assert not "Now creating plots" in res.output  # evaluation not attempted

    # Again, but with the advanced model
    res = invoke_cli(("run", ADVANCED_MODEL, "-d"))
    _check_result(res, expected_exit=0)

    # Adjusting some of the meta config parameters, testing if they show up
    args = ("run", DUMMY_MODEL, "--no-eval", "--note", "some_note", "-d")
    res = invoke_cli(
        args + ("-N", "23", "--we", "5", "--ws", "7", "--mp", "foo.bar=ABCXYZ")
    )
    _check_result(res, expected_exit=0)

    assert "Updates to meta configuration" in res.output
    assert "ABCXYZ" in res.output


def test_eval(with_test_models, tmp_output_dir, delay):
    """Tests the invocation of the utopya eval command"""
    # Simplest case
    res = invoke_cli(("eval", ADVANCED_MODEL, "-d"))
    _check_result(res, expected_exit=0)

    time.sleep(1)

    # Adjusting some of the meta config parameters
    args = ("eval", ADVANCED_MODEL, "-d")
    res = invoke_cli(args + ("-p", "plot_manager.raise_exc=true"))
    _check_result(res, expected_exit=0)
    assert "Updates to meta configuration" in res.output



"""Implements basic generic plot universe plot functions"""

import logging
from typing import Tuple, Union

from .. import DataManager, UniverseGroup
from . import PlotHelper, UniversePlotCreator, is_plot_func

log = logging.getLogger(__name__)

# -----------------------------------------------------------------------------


@is_plot_func(creator_type=UniversePlotCreator)
def lineplot(
    dm: DataManager,
    *,
    uni: UniverseGroup,
    hlpr: PlotHelper,
    model_name: str,
    path_to_data: Union[str, Tuple[str, str]],
    **plot_kwargs,
):
    """Performs an errorbar plot of a specific universe.

    Args:
        dm (DataManager): The data manager from which to retrieve the data
        uni (UniverseGroup): The data for this universe
        hlpr (PlotHelper): The PlotHelper
        model_name (str): The name of the model the data resides in
        path_to_data (str or Tuple[str, str]): The path to the data within the
            model data or the paths to the x and the y data, respectively
        **plot_kwargs: Passed on to plt.plot

    Raises:
        ValueError: On invalid data dimensionality
        ValueError: On mismatch of data shapes
    """
    # Get the data
    if isinstance(path_to_data, str):
        data_y = uni["data"][model_name][path_to_data]
        data_x = data_y.coords[data_y.dims[0]]

    else:
        data_x = uni["data"][model_name][path_to_data[0]]
        data_y = uni["data"][model_name][path_to_data[1]]

        if data_x.shape != data_y.shape:
            raise ValueError(
                "Mismatch of data shapes! "
                f"Were {data_x.shape} and {data_y.shape}, but have to be of "
                "same shape (after transfromation)."
            )

    # Require 1D data now
    if data_x.ndim != 1:
        raise ValueError(
            f"Lineplot requires 1D data, but got {data_x.ndim}D data of "
            f"shape {data_x.shape} for x-axis data:\n{data_x}\n"
            "Apply dimensionality reducing transformations to arrive "
            "at plottable data."
        )

    if data_y.ndim != 1:
        raise ValueError(
            f"Lineplot requires 1D data, but got {data_y.ndim}D data of "
            f"shape {data_y.shape} for y-axis data:\n{data_y}\n"
            "Apply dimensionality reducing transformations to arrive "
            "at plottable data."
        )

    hlpr.ax.plot(data_x, data_y, **plot_kwargs)



"""DAG-based distribution plotting functions"""

import copy
import logging
from typing import Sequence, Union

import xarray as xr

from . import PlotHelper, is_plot_func

log = logging.getLogger(__name__)


# -----------------------------------------------------------------------------


@is_plot_func(
    use_dag=True,
    required_dag_tags=("counts",),
    supports_animation=True,
    helper_defaults=dict(
        set_legend=dict(use_legend=True, loc="best", fontsize="small"),
        set_labels=dict(y="Counts"),
    ),
)
def histogram(
    *,
    data: dict,
    hlpr: PlotHelper,
    x: str,
    hue: str = None,
    frames: str = None,
    coarsen_by: int = None,
    align: str = "edge",
    bin_widths: Union[str, Sequence[float]] = None,
    suptitle_kwargs: dict = None,
    show_histogram_info: bool = True,
    **bar_kwargs,
):
    """Shows a distribution as a stacked bar plot, allowing animation.

    Expects as DAG result ``counts`` an xr.DataArray of one, two, or three
    dimensions. Depending on the ``hue`` and ``frames`` arguments, this will
    be represented as a stacked barplot and as an animation, respectively.

    Args:
        data (dict): The DAG results
        hlpr (PlotHelper): The PlotHelper
        x (str): The name of the dimension that represents the position of the
            histogram bins. By default, these are the bin *centers*.
        hue (str, optional): Which dimension to represent by stacking bars of
            different hue on top of each other
        frames (str, optional): Which dimension to represent by animation
        coarsen_by (int, optional): By which factor to coarsen the dimension
            specified by ``x``. Uses xr.DataArray.coarsen and pads boundary
            values.
        align (str, optional): Where to align bins. By default, uses ``edge``
            for alignment, as this is more exact for histograms.
        bin_widths (Union[str, Sequence[float]], optional): If not given, will
            use the difference between the ``x`` coordinates as bin widths,
            padding on the right side using the last value
            If a string, assume that it is a DAG result and retrieve it from
            ``data``. Otherwise, use it directly for the ``width`` argument of
            ``plt.bar``, i.e. assume it's a scalar or a sequence of bin widths.
        suptitle_kwargs (dict, optional): Description
        show_histogram_info (bool, optional): Whether to draw a box with
            information about the histogram.
        **bar_kwargs: Passed on ``hlpr.ax.bar`` invocation

    Returns:
        None

    Raises:
        ValueError: Bad dimensionality or missing ``bin_widths`` DAG result
    """

    def stacked_bar_plot(ax, dists: xr.DataArray, bin_widths):
        """Given a 2D xr.DataArray, plots a stacked barplot"""
        bottom = None  # to keep track of the bottom edges for stacking

        # Create the iterator
        if hue:
            hues = [c.item() for c in dists.coords[hue]]
            original_sorting = lambda c: hues.index(c[0])
            dist_iter = sorted(dists.groupby(hue), key=original_sorting)
        else:
            dist_iter = [(None, dists)]

        # Create the plots for each hue value
        for label, dist in dist_iter:
            dist = dist.squeeze(drop=True)
            ax.bar(
                dist.coords[x],
                dist,
                align=align,
                width=bin_widths,
                bottom=bottom,
                label=label,
                **bar_kwargs,
            )
            bottom = dist.data if bottom is None else bottom + dist.data

        # Annotate it
        if not show_histogram_info:
            return
        total_sum = dists.sum().item()
        hlpr.ax.text(
            1,
            1,
            (
                f"$N_{{bins}} = {dist.coords[x].size}$, "
                rf"$\Sigma_{{{x}}} = {total_sum:.4g}$"
            ),
            transform=hlpr.ax.transAxes,
            verticalalignment="bottom",
            horizontalalignment="right",
            fontdict=dict(fontsize="smaller"),
            bbox=dict(facecolor="white", linewidth=0.5, pad=2),
        )

    # Retrieve the data
    dists = data["counts"]

    # Check expected dimensions
    expected_ndim = 1 + bool(hue) + bool(frames)
    if dists.ndim != expected_ndim:
        raise ValueError(
            f"With `hue: {hue}` and `frames: {frames}`, expected "
            f"{expected_ndim}-dimensional data, but got:\n"
            f"{dists}"
        )

    # Calculate bin widths
    if bin_widths is None:
        bin_widths = dists.coords[x].diff(x)
        bin_widths = bin_widths.pad({x: (0, 1)}, mode="edge")

    elif isinstance(bin_widths, str):
        log.remark("Using DAG result '%s' for bin widths ...", bin_widths)
        try:
            bin_widths = data[bin_widths]
        except KeyError:
            raise ValueError(
                f"No DAG result '{bin_widths}' available for bin "
                "widths. Make sure `compute_only` is set such "
                "that the result will be computed."
            )

    # Allow dynamically plotting without animation
    if not frames:
        hlpr.disable_animation()
        stacked_bar_plot(hlpr.ax, dists, bin_widths)
        return
    # else: want an animation. Everything below here is only for that case.
    hlpr.enable_animation()

    # Determine the maximum, such that the scale is always the same
    max_counts = dists.sum(hue).max() if hue else dists.max()

    # Prepare some parameters for the update routine
    suptitle_kwargs = suptitle_kwargs if suptitle_kwargs else {}
    if "title" not in suptitle_kwargs:
        suptitle_kwargs["title"] = "{dim:} = {value:d}"

    # Define an animation update function. All frames are plotted therein.
    # There is no need to plot the first frame _outside_ the update function,
    # because it would be discarded anyway.
    def update():
        """The animation update function: a python generator"""
        log.note(
            "Commencing histogram animation for %d time steps ...",
            len(dists.coords[frames]),
        )

        for t, _dists in dists.groupby(frames):
            # Plot a frame onto an empty canvas
            hlpr.ax.clear()
            stacked_bar_plot(hlpr.ax, _dists, bin_widths)

            # Set the y-limits
            hlpr.invoke_helper("set_limits", y=[0, max_counts * 1.05])

            # Apply the suptitle format string, then invoke the helper
            st_kwargs = copy.deepcopy(suptitle_kwargs)
            st_kwargs["title"] = st_kwargs["title"].format(dim="time", value=t)
            hlpr.invoke_helper("set_suptitle", **st_kwargs)

            # Done with this frame. Let the writer grab it.
            yield

    # Register the animation update with the helper
    hlpr.register_animation_update(update, invoke_helpers_before_grab=True)



"""A generic, DAG-supporting time series plot"""

import xarray as xr

from . import PlotHelper, is_plot_func

# -----------------------------------------------------------------------------


@is_plot_func(
    use_dag=True,
    required_dag_tags=("y",),
    helper_defaults=dict(
        set_labels=dict(x="Time"), set_limits=dict(x=["min", "max"])
    ),
)
def time_series(
    *, data: dict, hlpr: PlotHelper, label_fstr: str = "{:.2g}", **plot_kwargs
):
    """This is a generic plotting function that plots one or multiple time
    series from the 'y' tag that is selected via the DAG framework.

    The y data needs to be an xarray object. If y is an xr.DataArray, it is
    assumed to be one- or two-dimensional.
    If y is an xr.Dataset, all data variables are plotted and their name is
    used as the label.

    For the x axis values, the corresponding 'time' coordinates are used.

    Args:
        data (dict): The data selected by the DAG framework
        hlpr (PlotHelper): The plot helper
        label_fstr (str, optional): Formatting to use for label
        **plot_kwargs: Passed on ot matplotlib.pyplot.plot
    """
    y = data["y"]

    # If this is an xr.DataArray, it may be one or two-dimensional
    if isinstance(y, xr.Dataset):
        # Simply plot all data variables
        for dvar, line in y.data_vars.items():
            hlpr.ax.plot(line.coords["time"], line, label=dvar, **plot_kwargs)

    elif isinstance(y, xr.DataArray):
        # Also allow two-dimensional arrays
        if y.ndim == 1:
            hlpr.ax.plot(y.coords["time"], y, **plot_kwargs)

        elif y.ndim == 2:
            loop_dim = [d for d in y.dims if d != "time"][0]

            for c in y.coords[loop_dim]:
                line = y.sel({loop_dim: c})
                hlpr.ax.plot(
                    line.coords["time"],
                    line,
                    label=label_fstr.format(c.item()),
                    **plot_kwargs,
                )

            # Provide a default title to the legend: name of the loop dimension
            hlpr.provide_defaults(
                "set_legend", title=f"${loop_dim}$ coordinate"
            )

        else:
            raise ValueError(
                "Given y-data needs to be of one- or two-dimensional, "
                f"was of dimensionality {y.ndim}! Data: {y}"
            )

    else:
        raise TypeError(f"Expected xr.Dataset or xr.DataArray, got {type(y)}")
